// Code generated by gocc; DO NOT EDIT.

package parser

const numNTSymbols = 111

type (
	gotoTable [numStates]gotoRow
	gotoRow   [numNTSymbols]int
)

var gotoTab = gotoTable{
	gotoRow{ // S0
		-1,  // S'
		1,   // Root
		2,   // StatementBlock
		3,   // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		4,   // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		5,   // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S1
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S5
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		168, // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		169, // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S6
		-1,  // S'
		-1,  // Root
		171, // StatementBlock
		3,   // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		4,   // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		5,   // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S7
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		174, // Rvalue
		204, // PrecedenceChainStart
		205, // TernaryTerm
		206, // LogicalOrTerm
		207, // LogicalXORTerm
		208, // LogicalAndTerm
		209, // AbsentCoalesceTerm
		210, // EmptyCoalesceTerm
		211, // EqneTerm
		212, // CmpTerm
		213, // BitwiseORTerm
		214, // BitwiseXORTerm
		215, // BitwiseANDTerm
		216, // BitwiseShiftTerm
		217, // AddsubdotTerm
		219, // MuldivTerm
		223, // DotTerm
		224, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		244, // MapLiteralKeyValuePairs
		245, // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S8
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S9
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S10
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S11
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S12
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S13
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S14
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S15
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S16
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S17
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S18
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S19
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S20
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S21
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S22
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S23
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S24
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S25
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S26
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S27
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S28
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S29
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		288, // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S30
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		292, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S31
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		382, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S32
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S33
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		385, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S34
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		476, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		475, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S35
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		481, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		480, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S36
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		482, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S37
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		483, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S38
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		485, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		484, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S39
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		304, // FieldValue
		299, // DirectFieldValue
		301, // IndirectFieldValue
		300, // BracedFieldValue
		305, // PositionalFieldName
		306, // PositionalFieldValue
		294, // FullSrec
		313, // OosvarValue
		296, // DirectOosvarValue
		298, // IndirectOosvarValue
		297, // BracedOosvarValue
		302, // FullOosvar
		293, // LocalVariable
		-1,  // Typedecl
		291, // Rvalue
		320, // PrecedenceChainStart
		321, // TernaryTerm
		322, // LogicalOrTerm
		323, // LogicalXORTerm
		324, // LogicalAndTerm
		325, // AbsentCoalesceTerm
		326, // EmptyCoalesceTerm
		327, // EqneTerm
		328, // CmpTerm
		329, // BitwiseORTerm
		330, // BitwiseXORTerm
		331, // BitwiseANDTerm
		332, // BitwiseShiftTerm
		333, // AddsubdotTerm
		335, // MuldivTerm
		339, // DotTerm
		340, // UnaryOpTerm
		343, // PowTerm
		344, // PrecedenceChainEnd
		345, // MlrvalOrFunction
		358, // ArrayLiteral
		-1,  // ArrayLiteralElements
		303, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		360, // ContextVariable
		373, // ENV
		375, // ArrayOrMapIndexAccess
		376, // ArrayOrMapPositionalNameAccess
		377, // ArrayOrMapPositionalValueAccess
		378, // ArraySliceAccess
		379, // FunctionCallsite
		380, // FunctionName
		486, // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		346, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S40
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		487, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S41
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S42
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		493, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		495, // EmittableList
		-1,  // EmittableAsList
		496, // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		500, // DirectFieldValue
		502, // IndirectFieldValue
		501, // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		494, // FullSrec
		-1,  // OosvarValue
		497, // DirectOosvarValue
		499, // IndirectOosvarValue
		498, // BracedOosvarValue
		503, // FullOosvar
		492, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		504, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S43
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		515, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S44
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		516, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		517, // EmittableAsList
		519, // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		500, // DirectFieldValue
		502, // IndirectFieldValue
		501, // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		494, // FullSrec
		-1,  // OosvarValue
		497, // DirectOosvarValue
		499, // IndirectOosvarValue
		498, // BracedOosvarValue
		503, // FullOosvar
		492, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		504, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S45
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		533, // FieldValue
		528, // DirectFieldValue
		530, // IndirectFieldValue
		529, // BracedFieldValue
		534, // PositionalFieldName
		535, // PositionalFieldValue
		523, // FullSrec
		542, // OosvarValue
		525, // DirectOosvarValue
		527, // IndirectOosvarValue
		526, // BracedOosvarValue
		531, // FullOosvar
		522, // LocalVariable
		-1,  // Typedecl
		521, // Rvalue
		549, // PrecedenceChainStart
		550, // TernaryTerm
		551, // LogicalOrTerm
		552, // LogicalXORTerm
		553, // LogicalAndTerm
		554, // AbsentCoalesceTerm
		555, // EmptyCoalesceTerm
		556, // EqneTerm
		557, // CmpTerm
		558, // BitwiseORTerm
		559, // BitwiseXORTerm
		560, // BitwiseANDTerm
		561, // BitwiseShiftTerm
		562, // AddsubdotTerm
		564, // MuldivTerm
		568, // DotTerm
		569, // UnaryOpTerm
		572, // PowTerm
		573, // PrecedenceChainEnd
		574, // MlrvalOrFunction
		587, // ArrayLiteral
		-1,  // ArrayLiteralElements
		532, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		589, // ContextVariable
		602, // ENV
		604, // ArrayOrMapIndexAccess
		605, // ArrayOrMapPositionalNameAccess
		606, // ArrayOrMapPositionalValueAccess
		607, // ArraySliceAccess
		608, // FunctionCallsite
		609, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		575, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S46
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		611, // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		612, // EmittableAsList
		519, // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		500, // DirectFieldValue
		502, // IndirectFieldValue
		501, // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		494, // FullSrec
		-1,  // OosvarValue
		497, // DirectOosvarValue
		499, // IndirectOosvarValue
		498, // BracedOosvarValue
		503, // FullOosvar
		492, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		504, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S47
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S48
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S49
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S50
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S51
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S52
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S53
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S54
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S55
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S56
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S57
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S58
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S59
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		615, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S60
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S61
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		705, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S62
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		706, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S63
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S64
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S65
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S66
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		707, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S67
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S68
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S69
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S70
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S71
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S72
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S73
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S74
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S75
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S76
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S77
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S78
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S79
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S80
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S81
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S82
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S83
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S84
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S85
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S86
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S87
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S88
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S89
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S90
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S91
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S92
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S93
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S94
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		733, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S95
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S96
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		742, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S97
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		743, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S98
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		744, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S99
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S100
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S101
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		746, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S102
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		747, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S103
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S104
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S105
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S106
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S107
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S108
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S109
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S110
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S111
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S112
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S113
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S114
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S115
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S116
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S117
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S118
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S119
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		765, // FieldValue
		760, // DirectFieldValue
		762, // IndirectFieldValue
		761, // BracedFieldValue
		766, // PositionalFieldName
		767, // PositionalFieldValue
		755, // FullSrec
		775, // OosvarValue
		757, // DirectOosvarValue
		759, // IndirectOosvarValue
		758, // BracedOosvarValue
		763, // FullOosvar
		754, // LocalVariable
		-1,  // Typedecl
		753, // Rvalue
		782, // PrecedenceChainStart
		783, // TernaryTerm
		784, // LogicalOrTerm
		785, // LogicalXORTerm
		786, // LogicalAndTerm
		787, // AbsentCoalesceTerm
		788, // EmptyCoalesceTerm
		789, // EqneTerm
		790, // CmpTerm
		791, // BitwiseORTerm
		792, // BitwiseXORTerm
		793, // BitwiseANDTerm
		794, // BitwiseShiftTerm
		795, // AddsubdotTerm
		797, // MuldivTerm
		801, // DotTerm
		802, // UnaryOpTerm
		805, // PowTerm
		806, // PrecedenceChainEnd
		807, // MlrvalOrFunction
		820, // ArrayLiteral
		822, // ArrayLiteralElements
		764, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		823, // ContextVariable
		836, // ENV
		838, // ArrayOrMapIndexAccess
		839, // ArrayOrMapPositionalNameAccess
		840, // ArrayOrMapPositionalValueAccess
		841, // ArraySliceAccess
		842, // FunctionCallsite
		843, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		808, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S120
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S121
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S122
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S123
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S124
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S125
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S126
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S127
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S128
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S129
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S130
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S131
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S132
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S133
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S134
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S135
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S136
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S137
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S138
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S139
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S140
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S141
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		849, // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S142
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S143
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S144
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S145
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S146
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S147
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S148
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S149
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S150
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		850, // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S151
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		851, // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S152
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		853, // ElifBlock
		852, // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S153
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S154
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S155
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S156
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		858, // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		-1,  // FieldValue
		-1,  // DirectFieldValue
		-1,  // IndirectFieldValue
		-1,  // BracedFieldValue
		-1,  // PositionalFieldName
		-1,  // PositionalFieldValue
		-1,  // FullSrec
		-1,  // OosvarValue
		-1,  // DirectOosvarValue
		-1,  // IndirectOosvarValue
		-1,  // BracedOosvarValue
		-1,  // FullOosvar
		-1,  // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		-1,  // UnaryOpTerm
		-1,  // PowTerm
		-1,  // PrecedenceChainEnd
		-1,  // MlrvalOrFunction
		-1,  // ArrayLiteral
		-1,  // ArrayLiteralElements
		-1,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		-1,  // ContextVariable
		-1,  // ENV
		-1,  // ArrayOrMapIndexAccess
		-1,  // ArrayOrMapPositionalNameAccess
		-1,  // ArrayOrMapPositionalValueAccess
		-1,  // ArraySliceAccess
		-1,  // FunctionCallsite
		-1,  // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		-1,  // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S157
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S158
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S159
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S160
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S161
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S162
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S163
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S164
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S165
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S166
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		864, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S167
		-1,  // S'
		-1,  // Root
		865, // StatementBlock
		3,   // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		4,   // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		5,   // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S168
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S169
		-1,  // S'
		-1,  // Root
		867, // StatementBlock
		3,   // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		4,   // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		5,   // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S170
		-1,  // S'
		-1,  // Root
		868, // StatementBlock
		3,   // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		4,   // BracelessStatement
		8,   // Assignment
		9,   // Unset
		28,  // Lvalue
		10,  // BareBoolean
		11,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		12,  // PrintStatement
		13,  // PrintnStatement
		14,  // EprintStatement
		15,  // EprintnStatement
		16,  // DumpStatement
		17,  // EdumpStatement
		18,  // TeeStatement
		22,  // EmitFStatement
		19,  // Emit1Statement
		20,  // EmitStatement
		21,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		55,  // FieldValue
		50,  // DirectFieldValue
		52,  // IndirectFieldValue
		51,  // BracedFieldValue
		56,  // PositionalFieldName
		57,  // PositionalFieldValue
		41,  // FullSrec
		64,  // OosvarValue
		47,  // DirectOosvarValue
		49,  // IndirectOosvarValue
		48,  // BracedOosvarValue
		53,  // FullOosvar
		32,  // LocalVariable
		31,  // Typedecl
		29,  // Rvalue
		80,  // PrecedenceChainStart
		81,  // TernaryTerm
		82,  // LogicalOrTerm
		83,  // LogicalXORTerm
		84,  // LogicalAndTerm
		85,  // AbsentCoalesceTerm
		86,  // EmptyCoalesceTerm
		87,  // EqneTerm
		88,  // CmpTerm
		89,  // BitwiseORTerm
		90,  // BitwiseXORTerm
		91,  // BitwiseANDTerm
		92,  // BitwiseShiftTerm
		93,  // AddsubdotTerm
		95,  // MuldivTerm
		99,  // DotTerm
		100, // UnaryOpTerm
		103, // PowTerm
		104, // PrecedenceChainEnd
		105, // MlrvalOrFunction
		118, // ArrayLiteral
		-1,  // ArrayLiteralElements
		54,  // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		120, // ContextVariable
		133, // ENV
		135, // ArrayOrMapIndexAccess
		136, // ArrayOrMapPositionalNameAccess
		137, // ArrayOrMapPositionalValueAccess
		138, // ArraySliceAccess
		139, // FunctionCallsite
		140, // FunctionName
		-1,  // FcnArgs
		27,  // SubroutineCallsite
		-1,  // SubroutineName
		5,   // BracefulStatement
		142, // BeginBlock
		143, // EndBlock
		144, // CondBlock
		145, // IfChain
		152, // IfElifStar
		153, // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		146, // WhileLoop
		23,  // DoWhileLoop
		147, // ForLoop
		157, // ForLoopOneVariable
		158, // ForLoopTwoVariable
		159, // ForLoopMultivariable
		-1,  // MultiIndex
		160, // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		24,  // BreakStatement
		25,  // ContinueStatement
		148, // NamedFunctionDefinition
		106, // UnnamedFunctionDefinition
		149, // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		26,  // ReturnStatement
	},
	gotoRow{ // S171
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S172
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		174, // Rvalue
		204, // PrecedenceChainStart
		205, // TernaryTerm
		206, // LogicalOrTerm
		207, // LogicalXORTerm
		208, // LogicalAndTerm
		209, // AbsentCoalesceTerm
		210, // EmptyCoalesceTerm
		211, // EqneTerm
		212, // CmpTerm
		213, // BitwiseORTerm
		214, // BitwiseXORTerm
		215, // BitwiseANDTerm
		216, // BitwiseShiftTerm
		217, // AddsubdotTerm
		219, // MuldivTerm
		223, // DotTerm
		224, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		870, // MapLiteralKeyValuePairs
		245, // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S173
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S174
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S175
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S176
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S177
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		533, // FieldValue
		528, // DirectFieldValue
		530, // IndirectFieldValue
		529, // BracedFieldValue
		534, // PositionalFieldName
		535, // PositionalFieldValue
		523, // FullSrec
		542, // OosvarValue
		525, // DirectOosvarValue
		527, // IndirectOosvarValue
		526, // BracedOosvarValue
		531, // FullOosvar
		522, // LocalVariable
		-1,  // Typedecl
		872, // Rvalue
		549, // PrecedenceChainStart
		550, // TernaryTerm
		551, // LogicalOrTerm
		552, // LogicalXORTerm
		553, // LogicalAndTerm
		554, // AbsentCoalesceTerm
		555, // EmptyCoalesceTerm
		556, // EqneTerm
		557, // CmpTerm
		558, // BitwiseORTerm
		559, // BitwiseXORTerm
		560, // BitwiseANDTerm
		561, // BitwiseShiftTerm
		562, // AddsubdotTerm
		564, // MuldivTerm
		568, // DotTerm
		569, // UnaryOpTerm
		572, // PowTerm
		573, // PrecedenceChainEnd
		574, // MlrvalOrFunction
		587, // ArrayLiteral
		-1,  // ArrayLiteralElements
		532, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		589, // ContextVariable
		602, // ENV
		604, // ArrayOrMapIndexAccess
		605, // ArrayOrMapPositionalNameAccess
		606, // ArrayOrMapPositionalValueAccess
		607, // ArraySliceAccess
		608, // FunctionCallsite
		609, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		575, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S178
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S179
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S180
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S181
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S182
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S183
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S184
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S185
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S186
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S187
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S188
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S189
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S190
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		873, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S191
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S192
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		874, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S193
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		875, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S194
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S195
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S196
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S197
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		627, // FieldValue
		622, // DirectFieldValue
		624, // IndirectFieldValue
		623, // BracedFieldValue
		628, // PositionalFieldName
		629, // PositionalFieldValue
		617, // FullSrec
		636, // OosvarValue
		619, // DirectOosvarValue
		621, // IndirectOosvarValue
		620, // BracedOosvarValue
		625, // FullOosvar
		616, // LocalVariable
		-1,  // Typedecl
		876, // Rvalue
		643, // PrecedenceChainStart
		644, // TernaryTerm
		645, // LogicalOrTerm
		646, // LogicalXORTerm
		647, // LogicalAndTerm
		648, // AbsentCoalesceTerm
		649, // EmptyCoalesceTerm
		650, // EqneTerm
		651, // CmpTerm
		652, // BitwiseORTerm
		653, // BitwiseXORTerm
		654, // BitwiseANDTerm
		655, // BitwiseShiftTerm
		656, // AddsubdotTerm
		658, // MuldivTerm
		662, // DotTerm
		663, // UnaryOpTerm
		666, // PowTerm
		667, // PrecedenceChainEnd
		668, // MlrvalOrFunction
		681, // ArrayLiteral
		-1,  // ArrayLiteralElements
		626, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		683, // ContextVariable
		696, // ENV
		698, // ArrayOrMapIndexAccess
		699, // ArrayOrMapPositionalNameAccess
		700, // ArrayOrMapPositionalValueAccess
		701, // ArraySliceAccess
		702, // FunctionCallsite
		703, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		669, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S198
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S199
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S200
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S201
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S202
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S203
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S204
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S205
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S206
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S207
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S208
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S209
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S210
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S211
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S212
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S213
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S214
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S215
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S216
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S217
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S218
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		902, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S219
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S220
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		910, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S221
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		911, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S222
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		912, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S223
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S224
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S225
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		914, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S226
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		186, // FieldValue
		181, // DirectFieldValue
		183, // IndirectFieldValue
		182, // BracedFieldValue
		187, // PositionalFieldName
		188, // PositionalFieldValue
		176, // FullSrec
		195, // OosvarValue
		178, // DirectOosvarValue
		180, // IndirectOosvarValue
		179, // BracedOosvarValue
		184, // FullOosvar
		175, // LocalVariable
		-1,  // Typedecl
		-1,  // Rvalue
		-1,  // PrecedenceChainStart
		-1,  // TernaryTerm
		-1,  // LogicalOrTerm
		-1,  // LogicalXORTerm
		-1,  // LogicalAndTerm
		-1,  // AbsentCoalesceTerm
		-1,  // EmptyCoalesceTerm
		-1,  // EqneTerm
		-1,  // CmpTerm
		-1,  // BitwiseORTerm
		-1,  // BitwiseXORTerm
		-1,  // BitwiseANDTerm
		-1,  // BitwiseShiftTerm
		-1,  // AddsubdotTerm
		-1,  // MuldivTerm
		-1,  // DotTerm
		915, // UnaryOpTerm
		227, // PowTerm
		228, // PrecedenceChainEnd
		229, // MlrvalOrFunction
		242, // ArrayLiteral
		-1,  // ArrayLiteralElements
		185, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		246, // ContextVariable
		259, // ENV
		261, // ArrayOrMapIndexAccess
		262, // ArrayOrMapPositionalNameAccess
		263, // ArrayOrMapPositionalValueAccess
		264, // ArraySliceAccess
		265, // FunctionCallsite
		266, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		230, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S227
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S228
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S229
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S230
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S231
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S232
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S233
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S234
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S235
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S236
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S237
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S238
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S239
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S240
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S241
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S242
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S243
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		765, // FieldValue
		760, // DirectFieldValue
		762, // IndirectFieldValue
		761, // BracedFieldValue
		766, // PositionalFieldName
		767, // PositionalFieldValue
		755, // FullSrec
		775, // OosvarValue
		757, // DirectOosvarValue
		759, // IndirectOosvarValue
		758, // BracedOosvarValue
		763, // FullOosvar
		754, // LocalVariable
		-1,  // Typedecl
		753, // Rvalue
		782, // PrecedenceChainStart
		783, // TernaryTerm
		784, // LogicalOrTerm
		785, // LogicalXORTerm
		786, // LogicalAndTerm
		787, // AbsentCoalesceTerm
		788, // EmptyCoalesceTerm
		789, // EqneTerm
		790, // CmpTerm
		791, // BitwiseORTerm
		792, // BitwiseXORTerm
		793, // BitwiseANDTerm
		794, // BitwiseShiftTerm
		795, // AddsubdotTerm
		797, // MuldivTerm
		801, // DotTerm
		802, // UnaryOpTerm
		805, // PowTerm
		806, // PrecedenceChainEnd
		807, // MlrvalOrFunction
		820, // ArrayLiteral
		921, // ArrayLiteralElements
		764, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		823, // ContextVariable
		836, // ENV
		838, // ArrayOrMapIndexAccess
		839, // ArrayOrMapPositionalNameAccess
		840, // ArrayOrMapPositionalValueAccess
		841, // ArraySliceAccess
		842, // FunctionCallsite
		843, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		808, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S244
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S245
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S246
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S247
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S248
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S249
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S250
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S251
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S252
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S253
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S254
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S255
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S256
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S257
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S258
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S259
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S260
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S261
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S262
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S263
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S264
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S265
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S266
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S267
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S268
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		928, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S269
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		929, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S270
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		930, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S271
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		931, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S272
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		932, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S273
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		933, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S274
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		934, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S275
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		935, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S276
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		936, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S277
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		937, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S278
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		938, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S279
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		939, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S280
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		940, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S281
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		941, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S282
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		942, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S283
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		943, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S284
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		944, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S285
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		945, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S286
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		946, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S287
		-1,  // S'
		-1,  // Root
		-1,  // StatementBlock
		-1,  // NonEmptyStatementBlock
		-1,  // StatementBlockInBraces
		-1,  // BracelessStatement
		-1,  // Assignment
		-1,  // Unset
		-1,  // Lvalue
		-1,  // BareBoolean
		-1,  // FilterStatement
		-1,  // Redirector
		-1,  // RedirectTarget
		-1,  // PrintStatement
		-1,  // PrintnStatement
		-1,  // EprintStatement
		-1,  // EprintnStatement
		-1,  // DumpStatement
		-1,  // EdumpStatement
		-1,  // TeeStatement
		-1,  // EmitFStatement
		-1,  // Emit1Statement
		-1,  // EmitStatement
		-1,  // EmitPStatement
		-1,  // EmittableList
		-1,  // EmittableAsList
		-1,  // Emittable
		-1,  // EmitKeys
		397, // FieldValue
		392, // DirectFieldValue
		394, // IndirectFieldValue
		393, // BracedFieldValue
		398, // PositionalFieldName
		399, // PositionalFieldValue
		387, // FullSrec
		406, // OosvarValue
		389, // DirectOosvarValue
		391, // IndirectOosvarValue
		390, // BracedOosvarValue
		395, // FullOosvar
		386, // LocalVariable
		-1,  // Typedecl
		947, // Rvalue
		413, // PrecedenceChainStart
		414, // TernaryTerm
		415, // LogicalOrTerm
		416, // LogicalXORTerm
		417, // LogicalAndTerm
		418, // AbsentCoalesceTerm
		419, // EmptyCoalesceTerm
		420, // EqneTerm
		421, // CmpTerm
		422, // BitwiseORTerm
		423, // BitwiseXORTerm
		424, // BitwiseANDTerm
		425, // BitwiseShiftTerm
		426, // AddsubdotTerm
		428, // MuldivTerm
		432, // DotTerm
		433, // UnaryOpTerm
		436, // PowTerm
		437, // PrecedenceChainEnd
		438, // MlrvalOrFunction
		451, // ArrayLiteral
		-1,  // ArrayLiteralElements
		396, // MapLiteral
		-1,  // MapLiteralKeyValuePairs
		-1,  // MapLiteralKeyValuePair
		453, // ContextVariable
		466, // ENV
		468, // ArrayOrMapIndexAccess
		469, // ArrayOrMapPositionalNameAccess
		470, // ArrayOrMapPositionalValueAccess
		471, // ArraySliceAccess
		472, // FunctionCallsite
		473, // FunctionName
		-1,  // FcnArgs
		-1,  // SubroutineCallsite
		-1,  // SubroutineName
		-1,  // BracefulStatement
		-1,  // BeginBlock
		-1,  // EndBlock
		-1,  // CondBlock
		-1,  // IfChain
		-1,  // IfElifStar
		-1,  // IfBlock
		-1,  // ElifBlock
		-1,  // ElseBlock
		-1,  // WhileLoop
		-1,  // DoWhileLoop
		-1,  // ForLoop
		-1,  // ForLoopOneVariable
		-1,  // ForLoopTwoVariable
		-1,  // ForLoopMultivariable
		-1,  // MultiIndex
		-1,  // TripleForLoop
		-1,  // TripleForStart
		-1,  // TripleForContinuation
		-1,  // TripleForContinuationItem
		-1,  // TripleForUpdate
		-1,  // BreakStatement
		-1,  // ContinueStatement
		-1,  // NamedFunctionDefinition
		439, // UnnamedFunctionDefinition
		-1,  // SubroutineDefinition
		-1,  // FuncOrSubrParameterList
		-1,  // FuncOrSubrNonEmptyParameterList
		-1,  // FuncOrSubrParameter
		-1,  // UntypedFuncOrSubrParameterName
		-1,  // TypedFuncOrSubrParameterName
		-1,  // ReturnStatement
	},
	gotoRow{ // S288
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S289
		-1,   // S'
		-1,   // Root
		948,  // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S290
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1104, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S291
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S292
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S293
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S294
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S295
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1106, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S296
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S297
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S298
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S299
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S300
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S301
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S302
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S303
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S304
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S305
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S306
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S307
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S308
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1107, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S309
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S310
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1108, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S311
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1109, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S312
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S313
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S314
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S315
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1110, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S316
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S317
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S318
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S319
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S320
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S321
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S322
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S323
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S324
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S325
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S326
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S327
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S328
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S329
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S330
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S331
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S332
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S333
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S334
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1136, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S335
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S336
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1144, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S337
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1145, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S338
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1146, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S339
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S340
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S341
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1148, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S342
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1149, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S343
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S344
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S345
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S346
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S347
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S348
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S349
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S350
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S351
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S352
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S353
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S354
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S355
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S356
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S357
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S358
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S359
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		1155, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S360
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S361
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S362
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S363
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S364
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S365
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S366
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S367
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S368
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S369
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S370
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S371
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S372
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S373
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S374
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S375
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S376
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S377
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S378
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S379
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S380
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S381
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S382
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S383
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S384
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1161, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S385
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S386
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S387
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S388
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1162, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S389
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S390
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S391
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S392
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S393
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S394
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S395
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S396
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S397
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S398
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S399
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S400
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S401
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1163, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S402
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S403
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1164, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S404
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1165, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S405
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S406
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S407
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S408
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1166, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S409
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S410
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S411
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S412
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S413
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S414
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S415
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S416
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S417
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S418
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S419
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S420
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S421
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S422
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S423
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S424
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S425
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S426
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S427
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1192, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S428
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S429
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1200, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S430
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1201, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S431
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1202, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S432
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S433
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S434
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1204, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S435
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1205, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S436
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S437
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S438
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S439
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S440
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S441
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S442
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S443
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S444
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S445
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S446
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S447
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S448
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S449
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S450
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S451
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S452
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		1211, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S453
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S454
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S455
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S456
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S457
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S458
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S459
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S460
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S461
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S462
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S463
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S464
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S465
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S466
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S467
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S468
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S469
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S470
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S471
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S472
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S473
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S474
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S475
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S476
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S477
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1218, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		1217, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S478
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1221, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		1217, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S479
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1222, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		1217, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S480
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S481
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S482
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S483
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S484
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S485
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S486
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S487
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S488
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1229, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		1227, // Rvalue
		1258, // PrecedenceChainStart
		1259, // TernaryTerm
		1260, // LogicalOrTerm
		1261, // LogicalXORTerm
		1262, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S489
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1320, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		1227, // Rvalue
		1258, // PrecedenceChainStart
		1259, // TernaryTerm
		1260, // LogicalOrTerm
		1261, // LogicalXORTerm
		1262, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S490
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		1321, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		1227, // Rvalue
		1258, // PrecedenceChainStart
		1259, // TernaryTerm
		1260, // LogicalOrTerm
		1261, // LogicalXORTerm
		1262, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S491
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1323, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S492
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S493
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S494
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S495
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S496
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S497
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S498
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S499
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S500
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S501
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S502
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S503
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S504
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S505
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S506
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1326, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S507
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S508
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S509
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S510
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1327, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S511
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S512
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S513
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S514
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S515
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S516
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S517
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S518
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		1333, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S519
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S520
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1354, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S521
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S522
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S523
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S524
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1356, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S525
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S526
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S527
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S528
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S529
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S530
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S531
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S532
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S533
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S534
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S535
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S536
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S537
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1357, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S538
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S539
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1358, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S540
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1359, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S541
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S542
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S543
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S544
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1360, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S545
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S546
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S547
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S548
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S549
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S550
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S551
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S552
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S553
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S554
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S555
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S556
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S557
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S558
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S559
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S560
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S561
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S562
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S563
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1386, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S564
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S565
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1394, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S566
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1395, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S567
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1396, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S568
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S569
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S570
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1398, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S571
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1399, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S572
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S573
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S574
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S575
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S576
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S577
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S578
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S579
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S580
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S581
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S582
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S583
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S584
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S585
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S586
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S587
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S588
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		1405, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S589
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S590
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S591
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S592
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S593
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S594
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S595
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S596
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S597
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S598
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S599
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S600
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S601
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S602
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S603
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S604
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S605
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S606
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S607
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S608
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S609
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S610
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S611
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S612
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S613
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		1412, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S614
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1414, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S615
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S616
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S617
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S618
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1416, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S619
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S620
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S621
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S622
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S623
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S624
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S625
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S626
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S627
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S628
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S629
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S630
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S631
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1417, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S632
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S633
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1418, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S634
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1419, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S635
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S636
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S637
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S638
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1420, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S639
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S640
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S641
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S642
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S643
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S644
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S645
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S646
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S647
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S648
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S649
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S650
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S651
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S652
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S653
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S654
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S655
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S656
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S657
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1446, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S658
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S659
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1454, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S660
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1455, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S661
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1456, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S662
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S663
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S664
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1458, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S665
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1459, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S666
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S667
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S668
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S669
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S670
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S671
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S672
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S673
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S674
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S675
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S676
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S677
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S678
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S679
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S680
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S681
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S682
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		1465, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S683
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S684
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S685
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S686
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S687
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S688
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S689
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S690
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S691
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S692
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S693
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S694
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S695
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S696
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S697
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S698
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S699
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S700
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S701
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S702
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S703
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S704
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S705
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S706
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S707
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S708
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		1473, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S709
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		1474, // LogicalXORTerm
		84,   // LogicalAndTerm
		85,   // AbsentCoalesceTerm
		86,   // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S710
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		1475, // LogicalAndTerm
		85,   // AbsentCoalesceTerm
		86,   // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S711
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		1476, // AbsentCoalesceTerm
		86,   // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S712
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		1477, // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S713
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		1478, // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S714
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1479, // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S715
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1480, // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S716
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1481, // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S717
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1482, // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S718
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1483, // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S719
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1484, // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S720
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1485, // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S721
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1486, // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S722
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1487, // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S723
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		1488, // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S724
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		1489, // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S725
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		1490, // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S726
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1491, // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S727
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1492, // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S728
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1493, // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S729
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1494, // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S730
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1495, // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S731
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1496, // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S732
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1497, // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S733
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S734
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S735
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1498, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S736
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1499, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S737
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1500, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S738
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1501, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S739
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1502, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S740
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1503, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S741
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1504, // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S742
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S743
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S744
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S745
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1505, // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S746
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S747
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S748
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		1508, // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S749
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		1510, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S750
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1601, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S751
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1602, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S752
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		1604, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S753
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S754
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S755
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S756
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1606, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S757
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S758
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S759
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S760
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S761
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S762
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S763
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S764
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S765
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S766
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S767
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S768
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S769
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1607, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S770
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S771
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S772
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1608, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S773
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1609, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S774
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S775
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S776
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S777
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1610, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S778
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S779
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S780
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S781
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S782
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S783
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S784
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S785
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S786
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S787
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S788
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S789
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S790
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S791
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S792
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S793
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S794
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S795
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S796
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1636, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S797
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S798
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1644, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S799
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1645, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S800
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1646, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S801
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S802
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S803
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1648, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S804
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1649, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S805
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S806
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S807
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S808
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S809
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S810
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S811
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S812
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S813
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S814
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S815
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S816
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S817
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S818
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S819
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S820
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S821
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		1655, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S822
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S823
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S824
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S825
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S826
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S827
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S828
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S829
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S830
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S831
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S832
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S833
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S834
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S835
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S836
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S837
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S838
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S839
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S840
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S841
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S842
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S843
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S844
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S845
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S846
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		1662, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S847
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		1665, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S848
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S849
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S850
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S851
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S852
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S853
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S854
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S855
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		1758, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S856
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1759, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S857
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		1760, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S858
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S859
		-1,   // S'
		-1,   // Root
		1762, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S860
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		1764, // Assignment
		-1,   // Unset
		1765, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		1767, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		1855, // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S861
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		1861, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S862
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S863
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S864
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S865
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S866
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		1868, // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		4,    // BracelessStatement
		8,    // Assignment
		9,    // Unset
		28,   // Lvalue
		10,   // BareBoolean
		11,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		12,   // PrintStatement
		13,   // PrintnStatement
		14,   // EprintStatement
		15,   // EprintnStatement
		16,   // DumpStatement
		17,   // EdumpStatement
		18,   // TeeStatement
		22,   // EmitFStatement
		19,   // Emit1Statement
		20,   // EmitStatement
		21,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		31,   // Typedecl
		29,   // Rvalue
		80,   // PrecedenceChainStart
		81,   // TernaryTerm
		82,   // LogicalOrTerm
		83,   // LogicalXORTerm
		84,   // LogicalAndTerm
		85,   // AbsentCoalesceTerm
		86,   // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		27,   // SubroutineCallsite
		-1,   // SubroutineName
		5,    // BracefulStatement
		142,  // BeginBlock
		143,  // EndBlock
		144,  // CondBlock
		145,  // IfChain
		152,  // IfElifStar
		153,  // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		146,  // WhileLoop
		23,   // DoWhileLoop
		147,  // ForLoop
		157,  // ForLoopOneVariable
		158,  // ForLoopTwoVariable
		159,  // ForLoopMultivariable
		-1,   // MultiIndex
		160,  // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		24,   // BreakStatement
		25,   // ContinueStatement
		148,  // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		149,  // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		26,   // ReturnStatement
	},
	gotoRow{ // S867
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S868
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S869
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S870
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S871
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		1871, // Rvalue
		1899, // PrecedenceChainStart
		1900, // TernaryTerm
		1901, // LogicalOrTerm
		1902, // LogicalXORTerm
		1903, // LogicalAndTerm
		1904, // AbsentCoalesceTerm
		1905, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S872
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S873
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S874
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S875
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S876
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S877
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		1966, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S878
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		1967, // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S879
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		1968, // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S880
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		1969, // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S881
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		1970, // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S882
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		1971, // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S883
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1972, // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S884
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1973, // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S885
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1974, // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S886
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1975, // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S887
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		1976, // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S888
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1977, // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S889
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1978, // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S890
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1979, // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S891
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		1980, // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S892
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		1981, // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S893
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		1982, // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S894
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		1983, // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S895
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1984, // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S896
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1985, // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S897
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		1986, // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S898
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1987, // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S899
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1988, // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S900
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1989, // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S901
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		1990, // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S902
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S903
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1991, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S904
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1992, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S905
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1993, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S906
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1994, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S907
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1995, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S908
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1996, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S909
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		1997, // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S910
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S911
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S912
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S913
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		1998, // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S914
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S915
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S916
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2001, // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S917
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2002, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S918
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2004, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S919
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2005, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S920
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S921
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S922
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S923
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2007, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S924
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S925
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2009, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S926
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2010, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S927
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2012, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S928
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S929
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S930
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S931
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S932
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S933
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S934
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S935
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S936
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S937
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S938
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S939
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S940
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S941
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S942
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S943
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S944
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S945
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S946
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S947
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S948
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S949
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S950
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S951
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		2015, // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		2016, // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S952
		-1,   // S'
		-1,   // Root
		2018, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S953
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2020, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S954
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S955
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S956
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S957
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S958
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S959
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S960
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S961
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S962
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S963
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S964
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S965
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S966
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S967
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S968
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S969
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S970
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S971
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S972
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S973
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S974
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S975
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		2041, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S976
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2045, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S977
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S978
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2136, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S979
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2227, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2226, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S980
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2232, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2231, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S981
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2233, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S982
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2234, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S983
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2236, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2235, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S984
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		2237, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S985
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2238, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S986
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S987
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2241, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		2243, // EmittableList
		-1,   // EmittableAsList
		2244, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S988
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2263, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S989
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2264, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		2265, // EmittableAsList
		2267, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S990
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2268, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S991
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		2269, // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		2270, // EmittableAsList
		2267, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S992
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S993
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S994
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S995
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S996
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S997
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S998
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S999
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1000
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1001
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1002
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1003
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1004
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2272, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1005
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1006
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2273, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1007
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2274, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1008
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1009
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1010
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1011
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2275, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1012
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1013
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1014
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1015
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1016
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1017
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1018
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1019
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1020
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1021
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1022
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1023
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1024
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1025
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1026
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1027
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1028
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1029
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1030
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2301, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1031
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1032
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2310, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1033
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2311, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1034
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2312, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1035
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1036
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1037
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2314, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1038
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2315, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1039
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1040
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1041
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1042
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1043
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1044
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1045
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1046
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1047
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1048
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1049
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1050
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1051
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1052
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1053
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1054
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1055
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2321, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1056
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1057
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1058
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1059
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1060
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1061
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1062
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1063
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1064
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1065
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1066
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1067
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1068
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1069
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1070
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1071
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1072
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1073
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1074
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1075
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1076
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1077
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		2325, // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1078
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1079
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1080
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1081
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1082
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1083
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1084
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1085
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1086
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		2326, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1087
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		2327, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1088
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		2329, // ElifBlock
		2328, // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1089
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1090
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1091
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1092
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		2334, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1093
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1094
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1095
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1096
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1097
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1098
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1099
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1100
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1101
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1102
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2339, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1103
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1104
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1105
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		291,  // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		2341, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1106
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1107
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1108
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1109
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1110
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1111
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		2347, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1112
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		2348, // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1113
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		2349, // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1114
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		2350, // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1115
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		2351, // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1116
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		2352, // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1117
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2353, // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1118
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2354, // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1119
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2355, // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1120
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2356, // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1121
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2357, // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1122
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2358, // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1123
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2359, // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1124
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2360, // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1125
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2361, // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1126
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		2362, // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1127
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		2363, // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1128
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		2364, // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1129
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2365, // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1130
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2366, // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1131
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2367, // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1132
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2368, // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1133
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2369, // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1134
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2370, // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1135
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2371, // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1136
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1137
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2372, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1138
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2373, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1139
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2374, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1140
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2375, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1141
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2376, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1142
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2377, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1143
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2378, // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1144
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1145
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1146
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1147
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2379, // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1148
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1149
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1150
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2382, // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1151
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2383, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1152
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2385, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1153
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2386, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1154
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1155
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1156
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1157
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2389, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1158
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2390, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1159
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2392, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1160
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1161
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1162
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1163
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1164
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1165
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1166
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1167
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		2399, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1168
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		2400, // LogicalXORTerm
		417,  // LogicalAndTerm
		418,  // AbsentCoalesceTerm
		419,  // EmptyCoalesceTerm
		420,  // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1169
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		2401, // LogicalAndTerm
		418,  // AbsentCoalesceTerm
		419,  // EmptyCoalesceTerm
		420,  // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1170
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		2402, // AbsentCoalesceTerm
		419,  // EmptyCoalesceTerm
		420,  // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1171
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		2403, // EmptyCoalesceTerm
		420,  // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1172
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		2404, // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1173
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2405, // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1174
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2406, // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1175
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2407, // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1176
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2408, // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1177
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2409, // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1178
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2410, // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1179
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2411, // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1180
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2412, // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1181
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2413, // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1182
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		2414, // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1183
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		2415, // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1184
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		2416, // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1185
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2417, // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1186
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2418, // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1187
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2419, // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1188
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2420, // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1189
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2421, // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1190
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2422, // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1191
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2423, // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1192
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1193
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2424, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1194
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2425, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1195
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2426, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1196
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2427, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1197
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2428, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1198
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2429, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1199
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2430, // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1200
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1201
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1202
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1203
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2431, // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1204
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1205
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1206
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2434, // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1207
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2435, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1208
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2437, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1209
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2438, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1210
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1211
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1212
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1213
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2441, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1214
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2442, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1215
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2444, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1216
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		291,  // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		2445, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1217
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1218
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1219
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1220
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1221
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1222
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1223
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		291,  // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		2446, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1224
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		291,  // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		2447, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1225
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2448, // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1226
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2451, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1227
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1228
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1229
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1230
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1231
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1232
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1233
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2452, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1234
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1235
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1236
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1237
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1238
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1239
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1240
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1241
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1242
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1243
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1244
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1245
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1246
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2453, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1247
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1248
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2454, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1249
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2455, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1250
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1251
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1252
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1253
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2456, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1254
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1255
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1256
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1257
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1258
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1259
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1260
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1261
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1262
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1263
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1264
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1265
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1266
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1267
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1268
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1269
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1270
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1271
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1272
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2482, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1273
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1274
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2490, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1275
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2491, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1276
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2492, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1277
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1278
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1279
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2494, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1280
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2495, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1281
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1282
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1283
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1284
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1285
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1286
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1287
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1288
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1289
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1290
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1291
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1292
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1293
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1294
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1295
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1296
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1297
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2501, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1298
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1299
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1300
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1301
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1302
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1303
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1304
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1305
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1306
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1307
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1308
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1309
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1310
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1311
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1312
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1313
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1314
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1315
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1316
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1317
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1318
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1319
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1320
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1321
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1322
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1323
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1324
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		2507, // EmittableList
		-1,   // EmittableAsList
		496,  // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		500,  // DirectFieldValue
		502,  // IndirectFieldValue
		501,  // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		494,  // FullSrec
		-1,   // OosvarValue
		497,  // DirectOosvarValue
		499,  // IndirectOosvarValue
		498,  // BracedOosvarValue
		503,  // FullOosvar
		492,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		504,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1325
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		2508, // EmittableList
		-1,   // EmittableAsList
		496,  // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		500,  // DirectFieldValue
		502,  // IndirectFieldValue
		501,  // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		494,  // FullSrec
		-1,   // OosvarValue
		497,  // DirectOosvarValue
		499,  // IndirectOosvarValue
		498,  // BracedOosvarValue
		503,  // FullOosvar
		492,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		504,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1326
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1327
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1328
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		2511, // EmittableAsList
		519,  // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		500,  // DirectFieldValue
		502,  // IndirectFieldValue
		501,  // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		494,  // FullSrec
		-1,   // OosvarValue
		497,  // DirectOosvarValue
		499,  // IndirectOosvarValue
		498,  // BracedOosvarValue
		503,  // FullOosvar
		492,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		504,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1329
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		2514, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1330
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2516, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1331
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1332
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1333
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1334
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1335
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1336
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1337
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1338
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1339
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1340
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1341
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1342
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1343
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1344
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2519, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1345
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1346
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1347
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1348
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2520, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1349
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1350
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1351
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1352
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1353
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1354
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1355
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1356
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1357
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1358
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1359
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1360
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1361
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		2527, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1362
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		2528, // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1363
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		2529, // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1364
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		2530, // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1365
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		2531, // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1366
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		2532, // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1367
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2533, // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1368
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2534, // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1369
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2535, // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1370
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2536, // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1371
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2537, // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1372
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2538, // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1373
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2539, // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1374
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2540, // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1375
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2541, // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1376
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		2542, // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1377
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		2543, // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1378
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		2544, // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1379
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2545, // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1380
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2546, // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1381
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2547, // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1382
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2548, // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1383
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2549, // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1384
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2550, // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1385
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2551, // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1386
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1387
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2552, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1388
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2553, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1389
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2554, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1390
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2555, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1391
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2556, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1392
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2557, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1393
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2558, // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1394
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1395
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1396
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1397
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2559, // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1398
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1399
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1400
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2562, // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1401
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2563, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1402
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2565, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1403
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2566, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1404
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1405
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1406
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1407
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2569, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1408
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2570, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1409
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2572, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1410
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		2573, // EmittableAsList
		519,  // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		500,  // DirectFieldValue
		502,  // IndirectFieldValue
		501,  // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		494,  // FullSrec
		-1,   // OosvarValue
		497,  // DirectOosvarValue
		499,  // IndirectOosvarValue
		498,  // BracedOosvarValue
		503,  // FullOosvar
		492,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		504,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1411
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		2575, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1412
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1413
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1414
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1415
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1416
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1417
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1418
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1419
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1420
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1421
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		2583, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1422
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		2584, // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1423
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		2585, // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1424
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		2586, // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1425
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		2587, // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1426
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		2588, // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1427
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2589, // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1428
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2590, // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1429
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2591, // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1430
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2592, // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1431
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2593, // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1432
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2594, // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1433
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2595, // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1434
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2596, // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1435
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2597, // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1436
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		2598, // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1437
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		2599, // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1438
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		2600, // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1439
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2601, // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1440
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2602, // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1441
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2603, // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1442
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2604, // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1443
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2605, // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1444
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2606, // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1445
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2607, // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1446
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1447
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2608, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1448
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2609, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1449
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2610, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1450
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2611, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1451
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2612, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1452
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2613, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1453
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2614, // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1454
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1455
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1456
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1457
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2615, // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1458
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1459
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1460
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2618, // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1461
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2619, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1462
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2621, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1463
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2622, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1464
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1465
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1466
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1467
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2625, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1468
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2626, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1469
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2628, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1470
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1471
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1472
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1473
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1474
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1475
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1476
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1477
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1478
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1479
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1480
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1481
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1482
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1483
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1484
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1485
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1486
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1487
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1488
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1489
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1490
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1491
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1492
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1493
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1494
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1495
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1496
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1497
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1498
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1499
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1500
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1501
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1502
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1503
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1504
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1505
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1506
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2632, // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1507
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2633, // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1508
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1509
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2635, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1510
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1511
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1512
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1513
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2638, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1514
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1515
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1516
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1517
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1518
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1519
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1520
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1521
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1522
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1523
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1524
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1525
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1526
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2639, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1527
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1528
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2640, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1529
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2641, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1530
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1531
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1532
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1533
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2642, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1534
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1535
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1536
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1537
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1538
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1539
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1540
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1541
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2645, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1542
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1543
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1544
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1545
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1546
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1547
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1548
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1549
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1550
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1551
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1552
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1553
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2670, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1554
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1555
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2678, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1556
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2679, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1557
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2680, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1558
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1559
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1560
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2682, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1561
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2683, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1562
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1563
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1564
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1565
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1566
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1567
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1568
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1569
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1570
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1571
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1572
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1573
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1574
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1575
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1576
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1577
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1578
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2689, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1579
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1580
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1581
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1582
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1583
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1584
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1585
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1586
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1587
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1588
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1589
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1590
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1591
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1592
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1593
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1594
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1595
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1596
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1597
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1598
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1599
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1600
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1601
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1602
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1603
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1604
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1605
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2697, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1606
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1607
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1608
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1609
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1610
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1611
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		2703, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1612
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		2704, // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1613
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		2705, // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1614
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		2706, // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1615
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		2707, // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1616
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		2708, // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1617
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2709, // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1618
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2710, // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1619
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2711, // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1620
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2712, // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1621
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		2713, // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1622
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2714, // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1623
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2715, // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1624
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2716, // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1625
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		2717, // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1626
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		2718, // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1627
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		2719, // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1628
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		2720, // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1629
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2721, // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1630
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2722, // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1631
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		2723, // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1632
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2724, // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1633
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2725, // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1634
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2726, // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1635
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		2727, // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1636
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1637
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2728, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1638
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2729, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1639
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2730, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1640
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2731, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1641
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2732, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1642
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2733, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1643
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		2734, // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1644
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1645
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1646
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1647
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2735, // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1648
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1649
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1650
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2738, // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1651
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		2739, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1652
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2741, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1653
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2742, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1654
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1655
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1656
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1657
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1658
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2745, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1659
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2746, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1660
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2748, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1661
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1662
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1663
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2751, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1664
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1665
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1666
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1667
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1668
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2754, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1669
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1670
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1671
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1672
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1673
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1674
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1675
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1676
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1677
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1678
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1679
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1680
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1681
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1682
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2755, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1683
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1684
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2756, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1685
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2757, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1686
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1687
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1688
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1689
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2758, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1690
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1691
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1692
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1693
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1694
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1695
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1696
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1697
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1698
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1699
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1700
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1701
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1702
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1703
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1704
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1705
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1706
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1707
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1708
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2784, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1709
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1710
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2792, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1711
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2793, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1712
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2794, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1713
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1714
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1715
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2796, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1716
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2797, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1717
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1718
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1719
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1720
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1721
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1722
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1723
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1724
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1725
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1726
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1727
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1728
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1729
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1730
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1731
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1732
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1733
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2803, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1734
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1735
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1736
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1737
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1738
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1739
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1740
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1741
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1742
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1743
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1744
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1745
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1746
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1747
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1748
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1749
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1750
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1751
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1752
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1753
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1754
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1755
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1756
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		2808, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1757
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2810, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1758
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1759
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1760
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1761
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1762
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1763
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2816, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1764
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1765
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1766
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1767
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1768
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1769
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		2840, // LocalVariable
		-1,   // Typedecl
		2839, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		2841, // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1770
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1771
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1772
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1773
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1774
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1775
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1776
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1777
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1778
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1779
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1780
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1781
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1782
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2842, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1783
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1784
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2843, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1785
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2844, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1786
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1787
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1788
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1789
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2845, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1790
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1791
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1792
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1793
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1794
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1795
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1796
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1797
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1798
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1799
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1800
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1801
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1802
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1803
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1804
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1805
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1806
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1807
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1808
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2874, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1809
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1810
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2882, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1811
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2883, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1812
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2884, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1813
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1814
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1815
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2886, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1816
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2887, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1817
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1818
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1819
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1820
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1821
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1822
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1823
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1824
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1825
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1826
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1827
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1828
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1829
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1830
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1831
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1832
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1833
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2893, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1834
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1835
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1836
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1837
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1838
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1839
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1840
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1841
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1842
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1843
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1844
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1845
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1846
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1847
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1848
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1849
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1850
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1851
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1852
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1853
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1854
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1855
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1856
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1857
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		2900, // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1858
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1859
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1860
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1861
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1862
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1863
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1864
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1865
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1866
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2903, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1867
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		2904, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1868
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1869
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1870
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		2906, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1871
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1872
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1873
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1874
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2907, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1875
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1876
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1877
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1878
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1879
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1880
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1881
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1882
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1883
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1884
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1885
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1886
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1887
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2908, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1888
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1889
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2909, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1890
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2910, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1891
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1892
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1893
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1894
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2911, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1895
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1896
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1897
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1898
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1899
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1900
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1901
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1902
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1903
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1904
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1905
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1906
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1907
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1908
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1909
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1910
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1911
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1912
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1913
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2937, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1914
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1915
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2945, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1916
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2946, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1917
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2947, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1918
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1919
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1920
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2949, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1921
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		2950, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1922
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1923
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1924
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1925
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1926
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1927
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1928
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1929
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1930
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1931
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1932
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1933
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1934
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1935
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1936
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1937
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1938
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		2956, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S1939
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1940
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1941
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1942
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1943
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1944
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1945
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1946
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1947
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1948
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1949
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1950
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1951
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1952
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1953
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1954
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1955
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1956
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1957
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1958
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1959
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1960
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1961
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1962
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1963
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1964
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1965
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1966
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1967
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1968
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1969
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1970
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1971
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1972
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1973
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1974
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1975
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1976
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1977
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1978
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1979
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1980
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1981
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1982
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1983
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1984
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1985
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1986
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1987
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1988
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1989
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1990
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1991
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1992
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1993
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1994
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1995
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1996
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1997
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1998
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S1999
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2964, // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2000
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		2965, // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2001
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2002
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2003
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		2968, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2004
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2005
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2006
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2007
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2008
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2009
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2010
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2011
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2012
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2013
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2014
		-1,   // S'
		-1,   // Root
		2975, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S2015
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2016
		-1,   // S'
		-1,   // Root
		2977, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S2017
		-1,   // S'
		-1,   // Root
		2978, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S2018
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2019
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2020
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2021
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2980, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2022
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2981, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2023
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2982, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2024
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2983, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2025
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2984, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2026
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2985, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2027
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2986, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2028
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2987, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2029
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2988, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2030
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2989, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2031
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2990, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2032
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2991, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2033
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2992, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2034
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2993, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2035
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2994, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2036
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2995, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2037
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2996, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2038
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2997, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2039
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2998, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2040
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		2999, // Rvalue
		2164, // PrecedenceChainStart
		2165, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2041
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2042
		-1,   // S'
		-1,   // Root
		3000, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S2043
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		3002, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2044
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2045
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2046
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2047
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2048
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3004, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2049
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2050
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2051
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2052
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2053
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2054
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2055
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2056
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2057
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2058
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2059
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2060
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2061
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3005, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2062
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2063
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3006, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2064
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3007, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2065
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2066
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2067
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2068
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3008, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2069
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2070
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2071
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2072
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2073
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2074
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2075
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2076
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2077
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2078
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2079
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2080
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2081
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2082
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2083
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2084
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2085
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2086
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2087
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3034, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2088
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2089
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3042, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2090
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3043, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2091
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3044, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2092
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2093
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2094
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3046, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2095
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3047, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2096
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2097
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2098
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2099
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2100
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2101
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2102
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2103
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2104
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2105
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2106
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2107
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2108
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2109
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2110
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2111
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2112
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		3053, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2113
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2114
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2115
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2116
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2117
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2118
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2119
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2120
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2121
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2122
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2123
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2124
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2125
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2126
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2127
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2128
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2129
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2130
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2131
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2132
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2133
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2134
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2135
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		3059, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2136
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2137
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2138
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2139
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3060, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2140
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2141
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2142
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2143
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2144
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2145
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2146
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2147
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2148
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2149
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2150
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2151
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2152
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3061, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2153
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2154
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3062, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2155
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3063, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2156
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2157
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2158
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2159
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3064, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2160
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2161
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2162
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2163
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2164
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2165
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2166
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2167
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2168
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2169
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2170
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2171
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2172
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2173
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2174
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2175
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2176
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2177
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2178
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3090, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2179
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2180
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3098, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2181
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3099, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2182
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3100, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2183
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2184
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2185
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3102, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2186
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3103, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2187
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2188
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2189
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2190
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2191
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2192
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2193
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2194
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2195
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2196
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2197
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2198
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2199
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2200
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2201
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2202
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2203
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		3109, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2204
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2205
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2206
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2207
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2208
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2209
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2210
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2211
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2212
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2213
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2214
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2215
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2216
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2217
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2218
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2219
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2220
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2221
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2222
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2223
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2224
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2225
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2226
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2227
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2228
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		3116, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3115, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2229
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		3119, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3115, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2230
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		3120, // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3115, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2231
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2232
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2233
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2234
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2235
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2236
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2237
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2238
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2239
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		3125, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2240
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2241
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2242
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2243
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2244
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2245
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2246
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2247
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2248
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2249
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2250
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2251
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2252
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2253
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2254
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3128, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2255
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2256
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2257
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2258
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3129, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2259
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2260
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2261
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2262
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2263
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2264
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2265
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2266
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3132, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2267
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2268
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2269
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2270
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2271
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3136, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2272
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2273
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2274
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2275
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2276
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3141, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2277
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3142, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2278
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3143, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2279
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3144, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2280
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3145, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2281
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3146, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2282
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3147, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2283
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3148, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2284
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3149, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2285
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3150, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2286
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3151, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2287
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3152, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2288
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3153, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2289
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3154, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2290
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3155, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2291
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3156, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2292
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3157, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2293
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3158, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2294
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3159, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2295
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3160, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2296
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3161, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2297
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3162, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2298
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3163, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2299
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3164, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2300
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3165, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2301
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2302
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2303
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3166, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2304
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3167, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2305
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3168, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2306
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3169, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2307
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3170, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2308
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3171, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2309
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3172, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2310
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2311
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2312
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2313
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3173, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2314
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2315
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2316
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3176, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2317
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3177, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2318
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3179, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2319
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3180, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2320
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2321
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2322
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2323
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3183, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2324
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3184, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2325
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2326
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2327
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2328
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2329
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2330
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2331
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3188, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2332
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3189, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2333
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3190, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2334
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2335
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		1764, // Assignment
		-1,   // Unset
		1765, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		3192, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		3194, // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2336
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3195, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2337
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2338
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2339
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2340
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2341
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2342
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2343
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2344
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2345
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2346
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2347
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2348
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2349
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2350
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2351
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2352
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2353
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2354
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2355
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2356
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2357
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2358
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2359
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2360
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2361
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2362
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2363
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2364
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2365
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2366
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2367
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2368
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2369
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2370
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2371
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2372
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2373
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2374
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2375
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2376
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2377
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2378
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2379
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2380
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3201, // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2381
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3202, // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2382
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2383
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2384
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3205, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2385
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2386
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2387
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2388
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2389
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2390
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2391
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2392
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2393
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2394
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2395
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2396
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2397
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2398
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2399
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2400
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2401
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2402
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2403
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2404
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2405
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2406
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2407
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2408
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2409
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2410
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2411
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2412
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2413
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2414
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2415
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2416
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2417
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2418
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2419
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2420
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2421
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2422
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2423
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2424
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2425
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2426
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2427
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2428
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2429
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2430
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2431
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2432
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3215, // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2433
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3216, // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2434
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2435
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2436
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3219, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2437
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2438
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2439
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2440
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2441
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2442
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2443
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2444
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2445
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2446
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2447
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2448
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2449
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2450
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2451
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2452
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2453
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2454
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2455
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2456
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2457
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3232, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2458
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3233, // LogicalXORTerm
		1262, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2459
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3234, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2460
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3235, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2461
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3236, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2462
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3237, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2463
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3238, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2464
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3239, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2465
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3240, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2466
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3241, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2467
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3242, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2468
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3243, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2469
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3244, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2470
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3245, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2471
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3246, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2472
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3247, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2473
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3248, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2474
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3249, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2475
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3250, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2476
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3251, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2477
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3252, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2478
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3253, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2479
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3254, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2480
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3255, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2481
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3256, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2482
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2483
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3257, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2484
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3258, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2485
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3259, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2486
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3260, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2487
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3261, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2488
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3262, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2489
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3263, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2490
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2491
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2492
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2493
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3264, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2494
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2495
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2496
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3267, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2497
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3268, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2498
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3270, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2499
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3271, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2500
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2501
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2502
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2503
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3274, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2504
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3275, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2505
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3277, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2506
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2507
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2508
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2509
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2510
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2511
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2512
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3279, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2513
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2514
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2515
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2516
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2517
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2518
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3283, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2519
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2520
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2521
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2522
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2523
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2524
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2525
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2526
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2527
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2528
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2529
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2530
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2531
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2532
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2533
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2534
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2535
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2536
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2537
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2538
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2539
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2540
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2541
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2542
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2543
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2544
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2545
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2546
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2547
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2548
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2549
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2550
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2551
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2552
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2553
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2554
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2555
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2556
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2557
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2558
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2559
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2560
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3289, // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2561
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3290, // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2562
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2563
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2564
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3293, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2565
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2566
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2567
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2568
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2569
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2570
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2571
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2572
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2573
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2574
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3301, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2575
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2576
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2577
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2578
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2579
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2580
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2581
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2582
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2583
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2584
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2585
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2586
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2587
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2588
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2589
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2590
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2591
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2592
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2593
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2594
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2595
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2596
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2597
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2598
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2599
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2600
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2601
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2602
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2603
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2604
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2605
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2606
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2607
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2608
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2609
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2610
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2611
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2612
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2613
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2614
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2615
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2616
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3306, // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2617
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3307, // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2618
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2619
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2620
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3310, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2621
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2622
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2623
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2624
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2625
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2626
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2627
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2628
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2629
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2630
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2631
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		55,   // FieldValue
		50,   // DirectFieldValue
		52,   // IndirectFieldValue
		51,   // BracedFieldValue
		56,   // PositionalFieldName
		57,   // PositionalFieldValue
		41,   // FullSrec
		64,   // OosvarValue
		47,   // DirectOosvarValue
		49,   // IndirectOosvarValue
		48,   // BracedOosvarValue
		53,   // FullOosvar
		32,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3318, // TernaryTerm
		82,   // LogicalOrTerm
		83,   // LogicalXORTerm
		84,   // LogicalAndTerm
		85,   // AbsentCoalesceTerm
		86,   // EmptyCoalesceTerm
		87,   // EqneTerm
		88,   // CmpTerm
		89,   // BitwiseORTerm
		90,   // BitwiseXORTerm
		91,   // BitwiseANDTerm
		92,   // BitwiseShiftTerm
		93,   // AddsubdotTerm
		95,   // MuldivTerm
		99,   // DotTerm
		100,  // UnaryOpTerm
		103,  // PowTerm
		104,  // PrecedenceChainEnd
		105,  // MlrvalOrFunction
		118,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		54,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		120,  // ContextVariable
		133,  // ENV
		135,  // ArrayOrMapIndexAccess
		136,  // ArrayOrMapPositionalNameAccess
		137,  // ArrayOrMapPositionalValueAccess
		138,  // ArraySliceAccess
		139,  // FunctionCallsite
		140,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		106,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2632
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2633
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2634
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2635
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2636
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2637
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3320, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2638
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2639
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2640
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2641
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2642
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2643
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3327, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2644
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3328, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2645
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2646
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2647
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3330, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2648
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3331, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2649
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3332, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2650
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3333, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2651
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3334, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2652
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3335, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2653
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3336, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2654
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3337, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2655
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3338, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2656
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3339, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2657
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3340, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2658
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3341, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2659
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3342, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2660
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3343, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2661
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3344, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2662
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3345, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2663
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3346, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2664
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3347, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2665
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3348, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2666
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3349, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2667
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3350, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2668
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3351, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2669
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3352, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2670
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2671
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3353, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2672
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3354, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2673
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3355, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2674
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3356, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2675
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3357, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2676
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3358, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2677
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3359, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2678
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2679
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2680
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2681
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3360, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2682
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2683
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2684
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3363, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2685
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3364, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2686
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3366, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2687
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3367, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2688
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2689
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2690
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2691
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3370, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2692
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3371, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2693
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3373, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2694
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2695
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2696
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2697
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2698
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2699
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2700
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2701
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2702
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2703
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2704
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2705
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2706
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2707
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2708
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2709
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2710
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2711
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2712
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2713
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2714
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2715
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2716
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2717
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2718
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2719
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2720
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2721
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2722
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2723
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2724
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2725
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2726
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2727
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2728
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2729
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2730
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2731
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2732
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2733
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2734
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2735
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2736
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3379, // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2737
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3380, // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2738
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2739
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2740
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3383, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2741
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2742
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2743
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2744
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2745
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2746
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2747
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2748
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2749
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2750
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2751
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2752
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3391, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2753
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2754
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2755
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2756
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2757
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2758
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2759
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3397, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2760
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3398, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2761
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3399, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2762
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3400, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2763
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3401, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2764
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3402, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2765
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3403, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2766
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3404, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2767
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3405, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2768
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3406, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2769
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3407, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2770
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3408, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2771
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3409, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2772
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3410, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2773
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3411, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2774
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3412, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2775
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3413, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2776
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3414, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2777
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3415, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2778
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3416, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2779
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3417, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2780
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3418, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2781
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3419, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2782
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3420, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2783
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3421, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2784
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2785
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3422, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2786
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3423, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2787
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3424, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2788
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3425, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2789
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3426, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2790
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3427, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2791
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3428, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2792
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2793
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2794
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2795
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3429, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2796
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2797
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2798
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3432, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2799
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3433, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2800
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3435, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2801
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3436, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2802
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2803
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2804
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2805
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3439, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2806
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3440, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2807
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3442, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2808
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2809
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2810
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2811
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3445, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2812
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3447, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2813
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3448, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2814
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2815
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2816
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2817
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3451, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2818
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3541, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2819
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3542, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2820
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3543, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2821
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3544, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2822
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3545, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2823
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3546, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2824
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3547, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2825
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3548, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2826
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3549, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2827
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3550, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2828
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3551, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2829
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3552, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2830
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3553, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2831
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3554, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2832
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3555, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2833
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3556, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2834
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3557, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2835
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3558, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2836
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		3559, // Rvalue
		3479, // PrecedenceChainStart
		3480, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2837
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		3560, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2838
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3562, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2839
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2840
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2841
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2842
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2843
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2844
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2845
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2846
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3571, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2847
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3572, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2848
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3573, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2849
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3574, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2850
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3575, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2851
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3576, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2852
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3577, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2853
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3578, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2854
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3579, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2855
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3580, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2856
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3581, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2857
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3582, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2858
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3583, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2859
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3584, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2860
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3585, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2861
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3586, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2862
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3587, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2863
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3588, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2864
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3589, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2865
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3590, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2866
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3591, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2867
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3592, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2868
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3593, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2869
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3594, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2870
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3595, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2871
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2872
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		2839, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2873
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2874
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2875
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3596, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2876
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3597, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2877
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3598, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2878
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3599, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2879
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3600, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2880
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3601, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2881
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3602, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2882
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2883
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2884
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2885
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3603, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2886
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2887
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2888
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3606, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2889
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3607, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2890
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3609, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2891
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3610, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2892
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2893
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2894
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2895
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3613, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2896
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3614, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2897
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		3617, // Assignment
		-1,   // Unset
		1765, // Lvalue
		3618, // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		31,   // Typedecl
		3619, // Rvalue
		3647, // PrecedenceChainStart
		3648, // TernaryTerm
		3649, // LogicalOrTerm
		3650, // LogicalXORTerm
		3651, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		3708, // TripleForContinuation
		3709, // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2898
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		3711, // Assignment
		-1,   // Unset
		1765, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2899
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3712, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2900
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2901
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3713, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2902
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		3716, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2903
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2904
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2905
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2906
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2907
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2908
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2909
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2910
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2911
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2912
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3725, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2913
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3726, // LogicalXORTerm
		1903, // LogicalAndTerm
		1904, // AbsentCoalesceTerm
		1905, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2914
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3727, // LogicalAndTerm
		1904, // AbsentCoalesceTerm
		1905, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2915
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3728, // AbsentCoalesceTerm
		1905, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2916
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3729, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2917
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3730, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2918
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3731, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2919
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3732, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2920
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3733, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2921
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3734, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2922
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3735, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2923
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3736, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2924
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3737, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2925
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3738, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2926
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3739, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2927
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3740, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2928
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3741, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2929
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3742, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2930
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3743, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2931
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3744, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2932
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3745, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2933
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3746, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2934
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3747, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2935
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3748, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2936
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3749, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2937
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2938
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3750, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2939
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3751, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2940
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3752, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2941
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3753, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2942
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3754, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2943
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3755, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2944
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3756, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2945
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2946
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2947
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2948
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3757, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2949
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2950
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2951
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3760, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2952
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3761, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2953
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3763, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2954
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3764, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2955
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2956
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2957
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2958
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3767, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2959
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3768, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2960
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3770, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2961
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2962
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2963
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3772, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2964
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2965
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2966
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2967
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3773, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2968
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2969
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2970
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2971
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2972
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2973
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2974
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3778, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S2975
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2976
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		3781, // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S2977
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2978
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2979
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2980
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2981
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2982
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2983
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2984
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2985
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2986
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2987
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2988
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2989
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2990
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2991
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2992
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2993
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2994
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2995
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2996
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2997
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2998
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S2999
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3000
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3001
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3002
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3003
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		3784, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3004
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3005
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3006
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3007
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3008
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3009
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3790, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3010
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3791, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3011
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3792, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3012
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3793, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3013
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3794, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3014
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3795, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3015
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3796, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3016
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3797, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3017
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3798, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3018
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3799, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3019
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3800, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3020
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3801, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3021
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3802, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3022
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3803, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3023
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3804, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3024
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3805, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3025
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3806, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3026
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3807, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3027
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3808, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3028
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3809, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3029
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3810, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3030
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3811, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3031
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3812, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3032
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3813, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3033
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3814, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3034
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3035
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3815, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3036
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3816, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3037
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3817, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3038
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3818, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3039
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3819, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3040
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3820, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3041
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3821, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3042
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3043
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3044
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3045
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3822, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3046
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3047
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3048
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3825, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3049
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3826, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3050
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3828, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3051
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3829, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3052
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3053
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3054
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3055
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3832, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3056
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3833, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3057
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3835, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3058
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3059
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3060
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3061
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3062
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3063
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3064
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3065
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3842, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3066
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		3843, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3067
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		3844, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3068
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		3845, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3069
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		3846, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3070
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		3847, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3071
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3848, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3072
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3849, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3073
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3850, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3074
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3851, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3075
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		3852, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3076
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3853, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3077
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3854, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3078
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3855, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3079
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		3856, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3080
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		3857, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3081
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		3858, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3082
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		3859, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3083
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3860, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3084
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3861, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3085
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		3862, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3086
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3863, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3087
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3864, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3088
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3865, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3089
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		3866, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3090
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3091
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3867, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3092
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3868, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3093
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3869, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3094
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3870, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3095
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3871, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3096
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3872, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3097
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		3873, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3098
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3099
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3100
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3101
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		3874, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3102
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3103
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3104
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3877, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3105
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		3878, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3106
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3880, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3107
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3881, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3108
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3109
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3110
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3111
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3884, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3112
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3885, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3113
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3887, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3114
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		3888, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3115
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3116
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3117
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3118
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3119
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3120
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3121
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		3889, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3122
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		2044, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		3890, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3123
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		3891, // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3124
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3125
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3126
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3894, // EmittableList
		-1,   // EmittableAsList
		2244, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3127
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		3895, // EmittableList
		-1,   // EmittableAsList
		2244, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3128
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3129
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3130
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		3898, // EmittableAsList
		2267, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3131
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3901, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3132
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3133
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3134
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		3903, // EmittableAsList
		2267, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		2248, // DirectFieldValue
		2250, // IndirectFieldValue
		2249, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		2242, // FullSrec
		-1,   // OosvarValue
		2245, // DirectOosvarValue
		2247, // IndirectOosvarValue
		2246, // BracedOosvarValue
		2251, // FullOosvar
		2240, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		2252, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3135
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3905, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3136
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3137
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3138
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3139
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3140
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3141
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3142
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3143
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3144
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3145
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3146
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3147
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3148
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3149
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3150
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3151
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3152
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3153
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3154
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3155
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3156
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3157
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3158
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3159
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3160
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3161
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3162
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3163
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3164
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3165
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3166
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3167
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3168
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3169
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3170
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3171
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3172
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3173
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3174
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3910, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3175
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3911, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3176
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3177
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3178
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3914, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3179
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3180
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3181
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3182
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3183
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3184
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3185
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3186
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		3920, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3187
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		3922, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3188
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3189
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3190
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3191
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3192
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3193
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		2840, // LocalVariable
		-1,   // Typedecl
		2839, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		3928, // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3194
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3195
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3196
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3931, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3197
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		3932, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3198
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3199
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3200
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3934, // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3201
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3202
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3203
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3204
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3935, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3205
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3206
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3207
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3208
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3209
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3210
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3211
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3940, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3212
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3213
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3214
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		397,  // FieldValue
		392,  // DirectFieldValue
		394,  // IndirectFieldValue
		393,  // BracedFieldValue
		398,  // PositionalFieldName
		399,  // PositionalFieldValue
		387,  // FullSrec
		406,  // OosvarValue
		389,  // DirectOosvarValue
		391,  // IndirectOosvarValue
		390,  // BracedOosvarValue
		395,  // FullOosvar
		386,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3944, // TernaryTerm
		415,  // LogicalOrTerm
		416,  // LogicalXORTerm
		417,  // LogicalAndTerm
		418,  // AbsentCoalesceTerm
		419,  // EmptyCoalesceTerm
		420,  // EqneTerm
		421,  // CmpTerm
		422,  // BitwiseORTerm
		423,  // BitwiseXORTerm
		424,  // BitwiseANDTerm
		425,  // BitwiseShiftTerm
		426,  // AddsubdotTerm
		428,  // MuldivTerm
		432,  // DotTerm
		433,  // UnaryOpTerm
		436,  // PowTerm
		437,  // PrecedenceChainEnd
		438,  // MlrvalOrFunction
		451,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		396,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		453,  // ContextVariable
		466,  // ENV
		468,  // ArrayOrMapIndexAccess
		469,  // ArrayOrMapPositionalNameAccess
		470,  // ArrayOrMapPositionalValueAccess
		471,  // ArraySliceAccess
		472,  // FunctionCallsite
		473,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		439,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3215
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3216
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3217
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3218
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3945, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3219
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3220
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3221
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3222
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3223
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3224
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3225
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3950, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3226
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3227
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3228
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3229
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3230
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3231
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3232
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3233
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3234
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3235
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3236
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3237
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3238
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3239
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3240
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3241
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3242
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3243
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3244
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3245
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3246
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3247
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3248
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3249
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3250
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3251
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3252
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3253
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3254
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3255
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3256
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3257
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3258
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3259
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3260
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3261
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3262
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3263
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3264
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3265
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3956, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3266
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3957, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3267
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3268
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3269
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3960, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3270
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3271
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3272
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3273
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3274
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3275
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3276
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3277
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3278
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3967, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3279
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3280
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3969, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3281
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3282
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3970, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3283
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3284
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3285
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3286
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3287
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3288
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3972, // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3289
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3290
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3291
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3292
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3973, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3293
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3294
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3295
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3296
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3297
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3298
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3299
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3978, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3300
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3981, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3301
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3302
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		3983, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3303
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3304
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3305
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		3985, // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3306
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3307
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3308
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3309
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		3986, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3310
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3311
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3312
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3313
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3314
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3315
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3316
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		3991, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3317
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3318
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3319
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3320
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3321
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3322
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3323
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3324
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3325
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3326
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3327
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3328
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3329
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3330
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3331
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3332
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3333
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3334
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3335
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3336
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3337
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3338
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3339
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3340
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3341
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3342
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3343
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3344
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3345
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3346
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3347
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3348
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3349
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3350
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3351
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3352
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3353
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3354
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3355
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3356
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3357
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3358
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3359
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3360
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3361
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3998, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3362
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		3999, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3363
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3364
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3365
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4002, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3366
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3367
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3368
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3369
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3370
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3371
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3372
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3373
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3374
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3375
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3376
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3377
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3378
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4011, // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3379
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3380
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3381
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3382
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4012, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3383
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3384
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3385
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3386
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3387
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3388
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3389
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4017, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3390
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3391
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3392
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3393
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3394
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3395
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3396
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3397
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3398
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3399
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3400
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3401
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3402
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3403
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3404
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3405
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3406
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3407
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3408
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3409
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3410
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3411
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3412
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3413
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3414
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3415
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3416
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3417
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3418
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3419
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3420
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3421
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3422
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3423
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3424
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3425
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3426
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3427
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3428
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3429
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3430
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4023, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3431
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4024, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3432
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3433
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3434
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4027, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3435
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3436
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3437
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3438
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3439
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3440
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3441
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3442
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3443
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3444
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4034, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3445
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3446
		-1,   // S'
		-1,   // Root
		4035, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3447
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3448
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3449
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3450
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		4038, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3451
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3452
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3453
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3454
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4039, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3455
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3456
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3457
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3458
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3459
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3460
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3461
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3462
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3463
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3464
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3465
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3466
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3467
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4040, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3468
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3469
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4041, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3470
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4042, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3471
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3472
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3473
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3474
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4043, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3475
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3476
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3477
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3478
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3479
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3480
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3481
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3482
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3483
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3484
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3485
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3486
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3487
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3488
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3489
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3490
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3491
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3492
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3493
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4069, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3494
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3495
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4077, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3496
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4078, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3497
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4079, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3498
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3499
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3500
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4081, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3501
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4082, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3502
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3503
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3504
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3505
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3506
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3507
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3508
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3509
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3510
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3511
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3512
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3513
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3514
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3515
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3516
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3517
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3518
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		4088, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3519
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3520
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3521
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3522
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3523
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3524
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3525
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3526
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3527
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3528
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3529
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3530
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3531
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3532
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3533
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3534
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3535
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3536
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3537
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3538
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3539
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3540
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3541
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3542
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3543
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3544
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3545
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3546
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3547
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3548
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3549
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3550
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3551
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3552
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3553
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3554
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3555
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3556
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3557
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3558
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3559
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3560
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3561
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3562
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3563
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3564
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		4095, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3565
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		4096, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3566
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3567
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3568
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3569
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3570
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3571
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3572
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3573
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3574
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3575
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3576
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3577
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3578
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3579
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3580
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3581
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3582
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3583
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3584
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3585
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3586
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3587
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3588
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3589
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3590
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3591
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3592
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3593
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3594
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3595
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3596
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3597
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3598
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3599
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3600
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3601
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3602
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3603
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3604
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4101, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3605
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4102, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3606
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3607
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3608
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4105, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3609
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3610
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3611
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3612
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3613
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3614
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3615
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3616
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		174,  // Rvalue
		204,  // PrecedenceChainStart
		205,  // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		4112, // MapLiteralKeyValuePairs
		245,  // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3617
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3618
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3619
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3620
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3621
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3622
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4113, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3623
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3624
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3625
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3626
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3627
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3628
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3629
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3630
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3631
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3632
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3633
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3634
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3635
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4114, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3636
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3637
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4115, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3638
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4116, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3639
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3640
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3641
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3642
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4117, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3643
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3644
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3645
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3646
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3647
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3648
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3649
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3650
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3651
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3652
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3653
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3654
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3655
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3656
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3657
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3658
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3659
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3660
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3661
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4143, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3662
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3663
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4151, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3664
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4152, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3665
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4153, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3666
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3667
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3668
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4155, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3669
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4156, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3670
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3671
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3672
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3673
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3674
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3675
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3676
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3677
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3678
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3679
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3680
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3681
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3682
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3683
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3684
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3685
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3686
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		765,  // FieldValue
		760,  // DirectFieldValue
		762,  // IndirectFieldValue
		761,  // BracedFieldValue
		766,  // PositionalFieldName
		767,  // PositionalFieldValue
		755,  // FullSrec
		775,  // OosvarValue
		757,  // DirectOosvarValue
		759,  // IndirectOosvarValue
		758,  // BracedOosvarValue
		763,  // FullOosvar
		754,  // LocalVariable
		-1,   // Typedecl
		753,  // Rvalue
		782,  // PrecedenceChainStart
		783,  // TernaryTerm
		784,  // LogicalOrTerm
		785,  // LogicalXORTerm
		786,  // LogicalAndTerm
		787,  // AbsentCoalesceTerm
		788,  // EmptyCoalesceTerm
		789,  // EqneTerm
		790,  // CmpTerm
		791,  // BitwiseORTerm
		792,  // BitwiseXORTerm
		793,  // BitwiseANDTerm
		794,  // BitwiseShiftTerm
		795,  // AddsubdotTerm
		797,  // MuldivTerm
		801,  // DotTerm
		802,  // UnaryOpTerm
		805,  // PowTerm
		806,  // PrecedenceChainEnd
		807,  // MlrvalOrFunction
		820,  // ArrayLiteral
		4162, // ArrayLiteralElements
		764,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		823,  // ContextVariable
		836,  // ENV
		838,  // ArrayOrMapIndexAccess
		839,  // ArrayOrMapPositionalNameAccess
		840,  // ArrayOrMapPositionalValueAccess
		841,  // ArraySliceAccess
		842,  // FunctionCallsite
		843,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		808,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3687
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3688
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3689
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3690
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3691
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3692
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3693
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3694
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3695
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3696
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3697
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3698
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3699
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3700
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3701
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3702
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3703
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3704
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3705
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3706
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3707
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3708
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3709
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3710
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3711
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3712
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3713
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3714
		-1,   // S'
		-1,   // Root
		4170, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3715
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4171, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3716
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3717
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4181, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3718
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4183, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3719
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3720
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3721
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3722
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3723
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3724
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3725
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3726
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3727
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3728
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3729
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3730
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3731
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3732
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3733
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3734
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3735
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3736
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3737
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3738
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3739
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3740
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3741
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3742
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3743
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3744
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3745
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3746
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3747
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3748
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3749
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3750
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3751
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3752
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3753
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3754
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3755
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3756
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3757
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3758
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4187, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3759
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4188, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3760
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3761
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3762
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4191, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3763
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3764
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3765
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3766
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3767
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3768
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3769
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3770
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3771
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3772
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3773
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3774
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3775
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3776
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3777
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3778
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3779
		-1,   // S'
		-1,   // Root
		4200, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3780
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4201, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3781
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3782
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3783
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3784
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3785
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3786
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3787
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3788
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3789
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3790
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3791
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3792
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3793
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3794
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3795
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3796
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3797
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3798
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3799
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3800
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3801
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3802
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3803
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3804
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3805
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3806
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3807
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3808
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3809
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3810
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3811
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3812
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3813
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3814
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3815
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3816
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3817
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3818
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3819
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3820
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3821
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3822
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3823
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4205, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3824
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4206, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3825
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3826
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3827
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4209, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3828
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3829
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3830
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3831
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3832
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3833
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3834
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3835
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3836
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3837
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3838
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3839
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3840
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3841
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3842
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3843
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3844
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3845
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3846
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3847
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3848
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3849
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3850
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3851
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3852
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3853
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3854
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3855
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3856
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3857
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3858
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3859
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3860
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3861
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3862
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3863
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3864
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3865
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3866
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3867
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3868
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3869
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3870
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3871
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3872
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3873
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3874
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3875
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4219, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3876
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4220, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3877
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3878
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3879
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4223, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3880
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3881
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3882
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3883
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3884
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3885
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3886
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3887
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3888
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3889
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3890
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3891
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3892
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3893
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3894
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3895
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3896
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3897
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3898
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3899
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		4231, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3900
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3901
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3902
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3903
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3904
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		4235, // EmittableList
		-1,   // EmittableAsList
		1334, // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		1338, // DirectFieldValue
		1340, // IndirectFieldValue
		1339, // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		1332, // FullSrec
		-1,   // OosvarValue
		1335, // DirectOosvarValue
		1337, // IndirectOosvarValue
		1336, // BracedOosvarValue
		1341, // FullOosvar
		1331, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		1342, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3905
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3906
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3907
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3908
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3909
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4238, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3910
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3911
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3912
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3913
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4239, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3914
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3915
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3916
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3917
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3918
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3919
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3920
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3921
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3922
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3923
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4246, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3924
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4248, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3925
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4249, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3926
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		4250, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3927
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4251, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3928
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3929
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		3617, // Assignment
		-1,   // Unset
		1765, // Lvalue
		3618, // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		31,   // Typedecl
		3619, // Rvalue
		3647, // PrecedenceChainStart
		3648, // TernaryTerm
		3649, // LogicalOrTerm
		3650, // LogicalXORTerm
		3651, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		4253, // TripleForContinuation
		3709, // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3930
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4254, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3931
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3932
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3933
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3934
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3935
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3936
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3937
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3938
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3939
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3940
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3941
		-1,   // S'
		-1,   // Root
		4261, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3942
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4262, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3943
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3944
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3945
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3946
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3947
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3948
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3949
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3950
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3951
		-1,   // S'
		-1,   // Root
		4265, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3952
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4266, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3953
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3954
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3955
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1242, // FieldValue
		1237, // DirectFieldValue
		1239, // IndirectFieldValue
		1238, // BracedFieldValue
		1243, // PositionalFieldName
		1244, // PositionalFieldValue
		1232, // FullSrec
		1251, // OosvarValue
		1234, // DirectOosvarValue
		1236, // IndirectOosvarValue
		1235, // BracedOosvarValue
		1240, // FullOosvar
		1228, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4268, // TernaryTerm
		1260, // LogicalOrTerm
		1261, // LogicalXORTerm
		1262, // LogicalAndTerm
		1263, // AbsentCoalesceTerm
		1264, // EmptyCoalesceTerm
		1265, // EqneTerm
		1266, // CmpTerm
		1267, // BitwiseORTerm
		1268, // BitwiseXORTerm
		1269, // BitwiseANDTerm
		1270, // BitwiseShiftTerm
		1271, // AddsubdotTerm
		1273, // MuldivTerm
		1277, // DotTerm
		1278, // UnaryOpTerm
		1281, // PowTerm
		1282, // PrecedenceChainEnd
		1283, // MlrvalOrFunction
		1296, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1241, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1298, // ContextVariable
		1311, // ENV
		1313, // ArrayOrMapIndexAccess
		1314, // ArrayOrMapPositionalNameAccess
		1315, // ArrayOrMapPositionalValueAccess
		1316, // ArraySliceAccess
		1317, // FunctionCallsite
		1318, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1284, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3956
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3957
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3958
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3959
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4269, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3960
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3961
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3962
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3963
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3964
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3965
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3966
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4274, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3967
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3968
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3969
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3970
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3971
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3972
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3973
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3974
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3975
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3976
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3977
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3978
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3979
		-1,   // S'
		-1,   // Root
		4280, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3980
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4281, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3981
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3982
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3983
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3984
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3985
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3986
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3987
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3988
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3989
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3990
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3991
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3992
		-1,   // S'
		-1,   // Root
		4285, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S3993
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4286, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3994
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3995
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3996
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3997
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4288, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S3998
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S3999
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4000
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4001
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4289, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4002
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4003
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4004
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4005
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4006
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4007
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4008
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4294, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4009
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4010
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4011
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4012
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4013
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4014
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4015
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4016
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4017
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4018
		-1,   // S'
		-1,   // Root
		4299, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4019
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4300, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4020
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4021
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4022
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4302, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4023
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4024
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4025
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4026
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4303, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4027
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4028
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4029
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4030
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4031
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4032
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4033
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4308, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4034
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4035
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4036
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4037
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4038
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4039
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4040
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4041
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4042
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4043
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4044
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4318, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4045
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		4319, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4046
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		4320, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4047
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		4321, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4048
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		4322, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4049
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		4323, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4050
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4324, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4051
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4325, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4052
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4326, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4053
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4327, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4054
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4328, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4055
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4329, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4056
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4330, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4057
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4331, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4058
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4332, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4059
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		4333, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4060
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		4334, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4061
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		4335, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4062
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4336, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4063
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4337, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4064
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4338, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4065
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4339, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4066
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4340, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4067
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4341, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4068
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4342, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4069
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4070
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4343, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4071
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4344, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4072
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4345, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4073
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4346, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4074
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4347, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4075
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4348, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4076
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4349, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4077
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4078
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4079
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4080
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4350, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4081
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4082
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4083
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4353, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4084
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		4354, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4085
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4356, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4086
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4357, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4087
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4088
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4089
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4090
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4360, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4091
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		4361, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4092
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		4363, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4093
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4364, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4094
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4365, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4095
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4096
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4097
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		4366, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4098
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4099
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4100
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4368, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4101
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4102
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4103
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4104
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4369, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4105
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4106
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4107
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4108
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4109
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4110
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4111
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4112
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4113
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4114
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4115
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4116
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4117
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4118
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		186,  // FieldValue
		181,  // DirectFieldValue
		183,  // IndirectFieldValue
		182,  // BracedFieldValue
		187,  // PositionalFieldName
		188,  // PositionalFieldValue
		176,  // FullSrec
		195,  // OosvarValue
		178,  // DirectOosvarValue
		180,  // IndirectOosvarValue
		179,  // BracedOosvarValue
		184,  // FullOosvar
		175,  // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4380, // TernaryTerm
		206,  // LogicalOrTerm
		207,  // LogicalXORTerm
		208,  // LogicalAndTerm
		209,  // AbsentCoalesceTerm
		210,  // EmptyCoalesceTerm
		211,  // EqneTerm
		212,  // CmpTerm
		213,  // BitwiseORTerm
		214,  // BitwiseXORTerm
		215,  // BitwiseANDTerm
		216,  // BitwiseShiftTerm
		217,  // AddsubdotTerm
		219,  // MuldivTerm
		223,  // DotTerm
		224,  // UnaryOpTerm
		227,  // PowTerm
		228,  // PrecedenceChainEnd
		229,  // MlrvalOrFunction
		242,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		185,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		246,  // ContextVariable
		259,  // ENV
		261,  // ArrayOrMapIndexAccess
		262,  // ArrayOrMapPositionalNameAccess
		263,  // ArrayOrMapPositionalValueAccess
		264,  // ArraySliceAccess
		265,  // FunctionCallsite
		266,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		230,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4119
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		4381, // LogicalXORTerm
		3651, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4120
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		4382, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4121
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		4383, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4122
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		4384, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4123
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		4385, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4124
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4386, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4125
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4387, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4126
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4388, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4127
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4389, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4128
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		4390, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4129
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4391, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4130
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4392, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4131
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4393, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4132
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		4394, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4133
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		4395, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4134
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		4396, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4135
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		4397, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4136
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4398, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4137
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4399, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4138
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		4400, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4139
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4401, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4140
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4402, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4141
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4403, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4142
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		4404, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4143
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4144
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4405, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4145
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4406, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4146
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4407, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4147
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4408, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4148
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4409, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4149
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4410, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4150
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		4411, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4151
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4152
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4153
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4154
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		4412, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4155
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4156
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4157
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4415, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4158
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1522, // FieldValue
		1517, // DirectFieldValue
		1519, // IndirectFieldValue
		1518, // BracedFieldValue
		1523, // PositionalFieldName
		1524, // PositionalFieldValue
		1512, // FullSrec
		1531, // OosvarValue
		1514, // DirectOosvarValue
		1516, // IndirectOosvarValue
		1515, // BracedOosvarValue
		1520, // FullOosvar
		1511, // LocalVariable
		-1,   // Typedecl
		4416, // Rvalue
		1538, // PrecedenceChainStart
		1539, // TernaryTerm
		1540, // LogicalOrTerm
		1542, // LogicalXORTerm
		1543, // LogicalAndTerm
		1544, // AbsentCoalesceTerm
		1545, // EmptyCoalesceTerm
		1546, // EqneTerm
		1547, // CmpTerm
		1548, // BitwiseORTerm
		1549, // BitwiseXORTerm
		1550, // BitwiseANDTerm
		1551, // BitwiseShiftTerm
		1552, // AddsubdotTerm
		1554, // MuldivTerm
		1558, // DotTerm
		1559, // UnaryOpTerm
		1562, // PowTerm
		1563, // PrecedenceChainEnd
		1564, // MlrvalOrFunction
		1577, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1521, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1579, // ContextVariable
		1592, // ENV
		1594, // ArrayOrMapIndexAccess
		1595, // ArrayOrMapPositionalNameAccess
		1596, // ArrayOrMapPositionalValueAccess
		1597, // ArraySliceAccess
		1598, // FunctionCallsite
		1599, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1565, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4159
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4418, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4160
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4419, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4161
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4162
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4163
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4164
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4422, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4165
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		1664, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		4423, // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4166
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		4425, // Assignment
		-1,   // Unset
		4426, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		4427, // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4167
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		3617, // Assignment
		-1,   // Unset
		1765, // Lvalue
		3618, // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		31,   // Typedecl
		3619, // Rvalue
		3647, // PrecedenceChainStart
		3648, // TernaryTerm
		3649, // LogicalOrTerm
		3650, // LogicalXORTerm
		3651, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		4428, // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4168
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		1857, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		4429, // FuncOrSubrParameterList
		1862, // FuncOrSubrNonEmptyParameterList
		1863, // FuncOrSubrParameter
		1864, // UntypedFuncOrSubrParameterName
		1865, // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4169
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4430, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4170
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4171
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4434, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4172
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4173
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4174
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4175
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4176
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4177
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4178
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4179
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4180
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4181
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4182
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4435, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4183
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4184
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4185
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4186
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1883, // FieldValue
		1878, // DirectFieldValue
		1880, // IndirectFieldValue
		1879, // BracedFieldValue
		1884, // PositionalFieldName
		1885, // PositionalFieldValue
		1873, // FullSrec
		1892, // OosvarValue
		1875, // DirectOosvarValue
		1877, // IndirectOosvarValue
		1876, // BracedOosvarValue
		1881, // FullOosvar
		1872, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4437, // TernaryTerm
		1901, // LogicalOrTerm
		1902, // LogicalXORTerm
		1903, // LogicalAndTerm
		1904, // AbsentCoalesceTerm
		1905, // EmptyCoalesceTerm
		1906, // EqneTerm
		1907, // CmpTerm
		1908, // BitwiseORTerm
		1909, // BitwiseXORTerm
		1910, // BitwiseANDTerm
		1911, // BitwiseShiftTerm
		1912, // AddsubdotTerm
		1914, // MuldivTerm
		1918, // DotTerm
		1919, // UnaryOpTerm
		1922, // PowTerm
		1923, // PrecedenceChainEnd
		1924, // MlrvalOrFunction
		1937, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1882, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1939, // ContextVariable
		1952, // ENV
		1954, // ArrayOrMapIndexAccess
		1955, // ArrayOrMapPositionalNameAccess
		1956, // ArrayOrMapPositionalValueAccess
		1957, // ArraySliceAccess
		1958, // FunctionCallsite
		1959, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1925, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4187
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4188
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4189
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4190
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4438, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4191
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4192
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4193
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4194
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4195
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4196
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4197
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4443, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4198
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4199
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4200
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4201
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4447, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4202
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4203
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4204
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4449, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4205
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4206
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4207
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4208
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4450, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4209
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4210
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4211
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4212
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4213
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4214
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4215
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4455, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4216
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4217
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4218
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		2148, // FieldValue
		2143, // DirectFieldValue
		2145, // IndirectFieldValue
		2144, // BracedFieldValue
		2149, // PositionalFieldName
		2150, // PositionalFieldValue
		2138, // FullSrec
		2157, // OosvarValue
		2140, // DirectOosvarValue
		2142, // IndirectOosvarValue
		2141, // BracedOosvarValue
		2146, // FullOosvar
		2137, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4459, // TernaryTerm
		2166, // LogicalOrTerm
		2167, // LogicalXORTerm
		2168, // LogicalAndTerm
		2169, // AbsentCoalesceTerm
		2170, // EmptyCoalesceTerm
		2171, // EqneTerm
		2172, // CmpTerm
		2173, // BitwiseORTerm
		2174, // BitwiseXORTerm
		2175, // BitwiseANDTerm
		2176, // BitwiseShiftTerm
		2177, // AddsubdotTerm
		2179, // MuldivTerm
		2183, // DotTerm
		2184, // UnaryOpTerm
		2187, // PowTerm
		2188, // PrecedenceChainEnd
		2189, // MlrvalOrFunction
		2202, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2147, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2204, // ContextVariable
		2217, // ENV
		2219, // ArrayOrMapIndexAccess
		2220, // ArrayOrMapPositionalNameAccess
		2221, // ArrayOrMapPositionalValueAccess
		2222, // ArraySliceAccess
		2223, // FunctionCallsite
		2224, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2190, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4219
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4220
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4221
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4222
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4460, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4223
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4224
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4225
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4226
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4227
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4228
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4229
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4465, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4230
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4468, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4231
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4232
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4470, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4233
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4471, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4234
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4472, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4235
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4236
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4474, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4237
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4238
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4239
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4240
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4241
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4242
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4243
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4244
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4245
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4477, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4246
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4247
		-1,   // S'
		-1,   // Root
		4478, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4248
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4249
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4250
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4251
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4252
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4253
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4254
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4255
		-1,   // S'
		-1,   // Root
		4484, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4256
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4485, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4257
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4486, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4258
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4488, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4259
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4260
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4261
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4262
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4490, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4263
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4264
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4265
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4266
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4492, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4267
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4268
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4269
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4270
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4271
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4272
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4273
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4274
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4275
		-1,   // S'
		-1,   // Root
		4495, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4276
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4496, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4277
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4497, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4278
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4279
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4280
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4281
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4499, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4282
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4500, // EmitKeys
		304,  // FieldValue
		299,  // DirectFieldValue
		301,  // IndirectFieldValue
		300,  // BracedFieldValue
		305,  // PositionalFieldName
		306,  // PositionalFieldValue
		294,  // FullSrec
		313,  // OosvarValue
		296,  // DirectOosvarValue
		298,  // IndirectOosvarValue
		297,  // BracedOosvarValue
		302,  // FullOosvar
		293,  // LocalVariable
		-1,   // Typedecl
		2513, // Rvalue
		320,  // PrecedenceChainStart
		321,  // TernaryTerm
		322,  // LogicalOrTerm
		323,  // LogicalXORTerm
		324,  // LogicalAndTerm
		325,  // AbsentCoalesceTerm
		326,  // EmptyCoalesceTerm
		327,  // EqneTerm
		328,  // CmpTerm
		329,  // BitwiseORTerm
		330,  // BitwiseXORTerm
		331,  // BitwiseANDTerm
		332,  // BitwiseShiftTerm
		333,  // AddsubdotTerm
		335,  // MuldivTerm
		339,  // DotTerm
		340,  // UnaryOpTerm
		343,  // PowTerm
		344,  // PrecedenceChainEnd
		345,  // MlrvalOrFunction
		358,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		303,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		360,  // ContextVariable
		373,  // ENV
		375,  // ArrayOrMapIndexAccess
		376,  // ArrayOrMapPositionalNameAccess
		377,  // ArrayOrMapPositionalValueAccess
		378,  // ArraySliceAccess
		379,  // FunctionCallsite
		380,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		346,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4283
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4284
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4285
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4286
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4502, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4287
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4288
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4289
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4290
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4291
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4292
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4293
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4294
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4295
		-1,   // S'
		-1,   // Root
		4505, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4296
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4506, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4297
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4298
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4299
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4300
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4508, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4301
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4302
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4303
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4304
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4305
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4306
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4307
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4308
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4309
		-1,   // S'
		-1,   // Root
		4511, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4310
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4512, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4311
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4312
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4313
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4314
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4315
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4316
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4317
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4318
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4319
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4320
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4321
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4322
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4323
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4324
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4325
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4326
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4327
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4328
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4329
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4330
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4331
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4332
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4333
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4334
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4335
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4336
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4337
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4338
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4339
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4340
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4341
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4342
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4343
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4344
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4345
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4346
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4347
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4348
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4349
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4350
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4351
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4516, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4352
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4517, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4353
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4354
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4355
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4520, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4356
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4357
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4358
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4359
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4360
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4361
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4362
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4363
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4364
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4365
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4366
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4367
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4368
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4369
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4370
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4371
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4372
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4373
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4374
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4375
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4376
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4377
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4378
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4379
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4380
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4381
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4382
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4383
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4384
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4385
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4386
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4387
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4388
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4389
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4390
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4391
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4392
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4393
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4394
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4395
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4396
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4397
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4398
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4399
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4400
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4401
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4402
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4403
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4404
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4405
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4406
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4407
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4408
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4409
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4410
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4411
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4412
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4413
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4534, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4414
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		4535, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4415
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4416
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4417
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4538, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4418
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4419
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4420
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4421
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4422
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4423
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4424
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4425
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4426
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4427
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4428
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4429
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4430
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4431
		-1,   // S'
		-1,   // Root
		4567, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4432
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4568, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4433
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4434
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4435
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4569, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4436
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4437
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4438
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4439
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4440
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4441
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4442
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4443
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4444
		-1,   // S'
		-1,   // Root
		4572, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4445
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4573, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4446
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4447
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4448
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4449
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4450
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4451
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4452
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4453
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4454
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4455
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4456
		-1,   // S'
		-1,   // Root
		4576, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4457
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4577, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4458
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4459
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4460
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4461
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4462
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4463
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4464
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4465
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4466
		-1,   // S'
		-1,   // Root
		4580, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4467
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4581, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4468
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4469
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4470
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4471
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4472
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4473
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4474
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4475
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4476
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4477
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4478
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4479
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4480
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4585, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4481
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4586, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4482
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		4587, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4483
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		4425, // Assignment
		-1,   // Unset
		4426, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		4588, // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4484
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4485
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4590, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4486
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4487
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4591, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4488
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4489
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4490
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4491
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4492
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4493
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4494
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4495
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4496
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4593, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4497
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4498
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4499
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4500
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4501
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4502
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4503
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4504
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4505
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4506
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4595, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4507
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4508
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4509
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4510
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4511
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4512
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4597, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4513
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4514
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4515
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3463, // FieldValue
		3458, // DirectFieldValue
		3460, // IndirectFieldValue
		3459, // BracedFieldValue
		3464, // PositionalFieldName
		3465, // PositionalFieldValue
		3453, // FullSrec
		3472, // OosvarValue
		3455, // DirectOosvarValue
		3457, // IndirectOosvarValue
		3456, // BracedOosvarValue
		3461, // FullOosvar
		3452, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4599, // TernaryTerm
		3481, // LogicalOrTerm
		3482, // LogicalXORTerm
		3483, // LogicalAndTerm
		3484, // AbsentCoalesceTerm
		3485, // EmptyCoalesceTerm
		3486, // EqneTerm
		3487, // CmpTerm
		3488, // BitwiseORTerm
		3489, // BitwiseXORTerm
		3490, // BitwiseANDTerm
		3491, // BitwiseShiftTerm
		3492, // AddsubdotTerm
		3494, // MuldivTerm
		3498, // DotTerm
		3499, // UnaryOpTerm
		3502, // PowTerm
		3503, // PrecedenceChainEnd
		3504, // MlrvalOrFunction
		3517, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3462, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3519, // ContextVariable
		3532, // ENV
		3534, // ArrayOrMapIndexAccess
		3535, // ArrayOrMapPositionalNameAccess
		3536, // ArrayOrMapPositionalValueAccess
		3537, // ArraySliceAccess
		3538, // FunctionCallsite
		3539, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3505, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4516
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4517
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4518
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4519
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4600, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4520
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4521
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4522
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4523
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4524
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4525
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4526
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4605, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4527
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4608, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4528
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4609, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4529
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4530
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4531
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4532
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4533
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		3631, // FieldValue
		3626, // DirectFieldValue
		3628, // IndirectFieldValue
		3627, // BracedFieldValue
		3632, // PositionalFieldName
		3633, // PositionalFieldValue
		3621, // FullSrec
		3640, // OosvarValue
		3623, // DirectOosvarValue
		3625, // IndirectOosvarValue
		3624, // BracedOosvarValue
		3629, // FullOosvar
		3620, // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		4611, // TernaryTerm
		3649, // LogicalOrTerm
		3650, // LogicalXORTerm
		3651, // LogicalAndTerm
		3652, // AbsentCoalesceTerm
		3653, // EmptyCoalesceTerm
		3654, // EqneTerm
		3655, // CmpTerm
		3656, // BitwiseORTerm
		3657, // BitwiseXORTerm
		3658, // BitwiseANDTerm
		3659, // BitwiseShiftTerm
		3660, // AddsubdotTerm
		3662, // MuldivTerm
		3666, // DotTerm
		3667, // UnaryOpTerm
		3670, // PowTerm
		3671, // PrecedenceChainEnd
		3672, // MlrvalOrFunction
		3685, // ArrayLiteral
		-1,   // ArrayLiteralElements
		3630, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		3687, // ContextVariable
		3700, // ENV
		3702, // ArrayOrMapIndexAccess
		3703, // ArrayOrMapPositionalNameAccess
		3704, // ArrayOrMapPositionalValueAccess
		3705, // ArraySliceAccess
		3706, // FunctionCallsite
		3707, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		3673, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4534
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4535
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4536
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4537
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		627,  // FieldValue
		622,  // DirectFieldValue
		624,  // IndirectFieldValue
		623,  // BracedFieldValue
		628,  // PositionalFieldName
		629,  // PositionalFieldValue
		617,  // FullSrec
		636,  // OosvarValue
		619,  // DirectOosvarValue
		621,  // IndirectOosvarValue
		620,  // BracedOosvarValue
		625,  // FullOosvar
		616,  // LocalVariable
		-1,   // Typedecl
		4612, // Rvalue
		643,  // PrecedenceChainStart
		644,  // TernaryTerm
		645,  // LogicalOrTerm
		646,  // LogicalXORTerm
		647,  // LogicalAndTerm
		648,  // AbsentCoalesceTerm
		649,  // EmptyCoalesceTerm
		650,  // EqneTerm
		651,  // CmpTerm
		652,  // BitwiseORTerm
		653,  // BitwiseXORTerm
		654,  // BitwiseANDTerm
		655,  // BitwiseShiftTerm
		656,  // AddsubdotTerm
		658,  // MuldivTerm
		662,  // DotTerm
		663,  // UnaryOpTerm
		666,  // PowTerm
		667,  // PrecedenceChainEnd
		668,  // MlrvalOrFunction
		681,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		626,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		683,  // ContextVariable
		696,  // ENV
		698,  // ArrayOrMapIndexAccess
		699,  // ArrayOrMapPositionalNameAccess
		700,  // ArrayOrMapPositionalValueAccess
		701,  // ArraySliceAccess
		702,  // FunctionCallsite
		703,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		669,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4538
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4539
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4540
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4541
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4542
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4543
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4544
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4617, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4545
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4618, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4546
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4619, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4547
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4620, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4548
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4621, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4549
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4622, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4550
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4623, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4551
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4624, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4552
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4625, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4553
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4626, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4554
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4627, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4555
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4628, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4556
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4629, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4557
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4630, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4558
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4631, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4559
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4632, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4560
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4633, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4561
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4634, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4562
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4635, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4563
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1678, // FieldValue
		1673, // DirectFieldValue
		1675, // IndirectFieldValue
		1674, // BracedFieldValue
		1679, // PositionalFieldName
		1680, // PositionalFieldValue
		1667, // FullSrec
		1687, // OosvarValue
		1670, // DirectOosvarValue
		1672, // IndirectOosvarValue
		1671, // BracedOosvarValue
		1676, // FullOosvar
		1666, // LocalVariable
		-1,   // Typedecl
		4636, // Rvalue
		1694, // PrecedenceChainStart
		1695, // TernaryTerm
		1696, // LogicalOrTerm
		1697, // LogicalXORTerm
		1698, // LogicalAndTerm
		1699, // AbsentCoalesceTerm
		1700, // EmptyCoalesceTerm
		1701, // EqneTerm
		1702, // CmpTerm
		1703, // BitwiseORTerm
		1704, // BitwiseXORTerm
		1705, // BitwiseANDTerm
		1706, // BitwiseShiftTerm
		1707, // AddsubdotTerm
		1709, // MuldivTerm
		1713, // DotTerm
		1714, // UnaryOpTerm
		1717, // PowTerm
		1718, // PrecedenceChainEnd
		1719, // MlrvalOrFunction
		1732, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1677, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1734, // ContextVariable
		1747, // ENV
		1749, // ArrayOrMapIndexAccess
		1750, // ArrayOrMapPositionalNameAccess
		1751, // ArrayOrMapPositionalValueAccess
		1752, // ArraySliceAccess
		1753, // FunctionCallsite
		1754, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1720, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4564
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		4637, // Assignment
		-1,   // Unset
		4426, // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1778, // FieldValue
		1773, // DirectFieldValue
		1775, // IndirectFieldValue
		1774, // BracedFieldValue
		1779, // PositionalFieldName
		1780, // PositionalFieldValue
		1768, // FullSrec
		1787, // OosvarValue
		1770, // DirectOosvarValue
		1772, // IndirectOosvarValue
		1771, // BracedOosvarValue
		1776, // FullOosvar
		2871, // LocalVariable
		31,   // Typedecl
		1766, // Rvalue
		1794, // PrecedenceChainStart
		1795, // TernaryTerm
		1796, // LogicalOrTerm
		1797, // LogicalXORTerm
		1798, // LogicalAndTerm
		1799, // AbsentCoalesceTerm
		1800, // EmptyCoalesceTerm
		1801, // EqneTerm
		1802, // CmpTerm
		1803, // BitwiseORTerm
		1804, // BitwiseXORTerm
		1805, // BitwiseANDTerm
		1806, // BitwiseShiftTerm
		1807, // AddsubdotTerm
		1809, // MuldivTerm
		1813, // DotTerm
		1814, // UnaryOpTerm
		1817, // PowTerm
		1818, // PrecedenceChainEnd
		1819, // MlrvalOrFunction
		1832, // ArrayLiteral
		-1,   // ArrayLiteralElements
		1777, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1834, // ContextVariable
		1847, // ENV
		1849, // ArrayOrMapIndexAccess
		1850, // ArrayOrMapPositionalNameAccess
		1851, // ArrayOrMapPositionalValueAccess
		1852, // ArraySliceAccess
		1853, // FunctionCallsite
		1854, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		1820, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4565
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4638, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4566
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4639, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4567
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4568
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4643, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4569
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4570
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4571
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4572
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4573
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4645, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4574
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4575
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4576
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4577
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4647, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4578
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4579
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4580
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4581
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4649, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4582
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4650, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4583
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		4651, // EmitKeys
		2057, // FieldValue
		2052, // DirectFieldValue
		2054, // IndirectFieldValue
		2053, // BracedFieldValue
		2058, // PositionalFieldName
		2059, // PositionalFieldValue
		2047, // FullSrec
		2066, // OosvarValue
		2049, // DirectOosvarValue
		2051, // IndirectOosvarValue
		2050, // BracedOosvarValue
		2055, // FullOosvar
		2046, // LocalVariable
		-1,   // Typedecl
		3900, // Rvalue
		2073, // PrecedenceChainStart
		2074, // TernaryTerm
		2075, // LogicalOrTerm
		2076, // LogicalXORTerm
		2077, // LogicalAndTerm
		2078, // AbsentCoalesceTerm
		2079, // EmptyCoalesceTerm
		2080, // EqneTerm
		2081, // CmpTerm
		2082, // BitwiseORTerm
		2083, // BitwiseXORTerm
		2084, // BitwiseANDTerm
		2085, // BitwiseShiftTerm
		2086, // AddsubdotTerm
		2088, // MuldivTerm
		2092, // DotTerm
		2093, // UnaryOpTerm
		2096, // PowTerm
		2097, // PrecedenceChainEnd
		2098, // MlrvalOrFunction
		2111, // ArrayLiteral
		-1,   // ArrayLiteralElements
		2056, // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		2113, // ContextVariable
		2126, // ENV
		2128, // ArrayOrMapIndexAccess
		2129, // ArrayOrMapPositionalNameAccess
		2130, // ArrayOrMapPositionalValueAccess
		2131, // ArraySliceAccess
		2132, // FunctionCallsite
		2133, // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		2099, // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4584
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4585
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4586
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4587
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4588
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4589
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4590
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4591
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4655, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4592
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4593
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4594
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4595
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4596
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4597
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4598
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4599
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4600
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4601
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4602
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4603
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4604
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4605
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4606
		-1,   // S'
		-1,   // Root
		4658, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4607
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4659, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4608
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4609
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4610
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4611
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4612
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4613
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4614
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4615
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4616
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4617
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4618
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4619
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4620
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4621
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4622
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4623
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4624
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4625
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4626
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4627
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4628
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4629
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4630
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4631
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4632
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4633
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4634
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4635
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4636
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4637
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4638
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4639
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4640
		-1,   // S'
		-1,   // Root
		4663, // StatementBlock
		949,  // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		950,  // BracelessStatement
		954,  // Assignment
		955,  // Unset
		974,  // Lvalue
		956,  // BareBoolean
		957,  // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		958,  // PrintStatement
		959,  // PrintnStatement
		960,  // EprintStatement
		961,  // EprintnStatement
		962,  // DumpStatement
		963,  // EdumpStatement
		964,  // TeeStatement
		968,  // EmitFStatement
		965,  // Emit1Statement
		966,  // EmitStatement
		967,  // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		1000, // FieldValue
		995,  // DirectFieldValue
		997,  // IndirectFieldValue
		996,  // BracedFieldValue
		1001, // PositionalFieldName
		1002, // PositionalFieldValue
		986,  // FullSrec
		1009, // OosvarValue
		992,  // DirectOosvarValue
		994,  // IndirectOosvarValue
		993,  // BracedOosvarValue
		998,  // FullOosvar
		977,  // LocalVariable
		31,   // Typedecl
		975,  // Rvalue
		1016, // PrecedenceChainStart
		1017, // TernaryTerm
		1018, // LogicalOrTerm
		1019, // LogicalXORTerm
		1020, // LogicalAndTerm
		1021, // AbsentCoalesceTerm
		1022, // EmptyCoalesceTerm
		1023, // EqneTerm
		1024, // CmpTerm
		1025, // BitwiseORTerm
		1026, // BitwiseXORTerm
		1027, // BitwiseANDTerm
		1028, // BitwiseShiftTerm
		1029, // AddsubdotTerm
		1031, // MuldivTerm
		1035, // DotTerm
		1036, // UnaryOpTerm
		1039, // PowTerm
		1040, // PrecedenceChainEnd
		1041, // MlrvalOrFunction
		1054, // ArrayLiteral
		-1,   // ArrayLiteralElements
		999,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		1056, // ContextVariable
		1069, // ENV
		1071, // ArrayOrMapIndexAccess
		1072, // ArrayOrMapPositionalNameAccess
		1073, // ArrayOrMapPositionalValueAccess
		1074, // ArraySliceAccess
		1075, // FunctionCallsite
		1076, // FunctionName
		-1,   // FcnArgs
		973,  // SubroutineCallsite
		-1,   // SubroutineName
		951,  // BracefulStatement
		1078, // BeginBlock
		1079, // EndBlock
		1080, // CondBlock
		1081, // IfChain
		1088, // IfElifStar
		1089, // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		1082, // WhileLoop
		969,  // DoWhileLoop
		1083, // ForLoop
		1093, // ForLoopOneVariable
		1094, // ForLoopTwoVariable
		1095, // ForLoopMultivariable
		-1,   // MultiIndex
		1096, // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		970,  // BreakStatement
		971,  // ContinueStatement
		1084, // NamedFunctionDefinition
		1042, // UnnamedFunctionDefinition
		1085, // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		972,  // ReturnStatement
	},
	gotoRow{ // S4641
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		4664, // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4642
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4643
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4644
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4645
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4646
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4647
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4648
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4649
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4650
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4651
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4652
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4665, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4653
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		-1,   // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		533,  // FieldValue
		528,  // DirectFieldValue
		530,  // IndirectFieldValue
		529,  // BracedFieldValue
		534,  // PositionalFieldName
		535,  // PositionalFieldValue
		523,  // FullSrec
		542,  // OosvarValue
		525,  // DirectOosvarValue
		527,  // IndirectOosvarValue
		526,  // BracedOosvarValue
		531,  // FullOosvar
		522,  // LocalVariable
		-1,   // Typedecl
		4666, // Rvalue
		549,  // PrecedenceChainStart
		550,  // TernaryTerm
		551,  // LogicalOrTerm
		552,  // LogicalXORTerm
		553,  // LogicalAndTerm
		554,  // AbsentCoalesceTerm
		555,  // EmptyCoalesceTerm
		556,  // EqneTerm
		557,  // CmpTerm
		558,  // BitwiseORTerm
		559,  // BitwiseXORTerm
		560,  // BitwiseANDTerm
		561,  // BitwiseShiftTerm
		562,  // AddsubdotTerm
		564,  // MuldivTerm
		568,  // DotTerm
		569,  // UnaryOpTerm
		572,  // PowTerm
		573,  // PrecedenceChainEnd
		574,  // MlrvalOrFunction
		587,  // ArrayLiteral
		-1,   // ArrayLiteralElements
		532,  // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		589,  // ContextVariable
		602,  // ENV
		604,  // ArrayOrMapIndexAccess
		605,  // ArrayOrMapPositionalNameAccess
		606,  // ArrayOrMapPositionalValueAccess
		607,  // ArraySliceAccess
		608,  // FunctionCallsite
		609,  // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		575,  // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4654
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4667, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4655
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4656
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4657
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4658
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4659
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4669, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4660
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4670, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4661
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4662
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4663
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4664
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4672, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4665
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4666
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4667
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4668
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4669
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4670
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4671
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4672
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
	gotoRow{ // S4673
		-1,   // S'
		-1,   // Root
		-1,   // StatementBlock
		-1,   // NonEmptyStatementBlock
		4674, // StatementBlockInBraces
		-1,   // BracelessStatement
		-1,   // Assignment
		-1,   // Unset
		-1,   // Lvalue
		-1,   // BareBoolean
		-1,   // FilterStatement
		-1,   // Redirector
		-1,   // RedirectTarget
		-1,   // PrintStatement
		-1,   // PrintnStatement
		-1,   // EprintStatement
		-1,   // EprintnStatement
		-1,   // DumpStatement
		-1,   // EdumpStatement
		-1,   // TeeStatement
		-1,   // EmitFStatement
		-1,   // Emit1Statement
		-1,   // EmitStatement
		-1,   // EmitPStatement
		-1,   // EmittableList
		-1,   // EmittableAsList
		-1,   // Emittable
		-1,   // EmitKeys
		-1,   // FieldValue
		-1,   // DirectFieldValue
		-1,   // IndirectFieldValue
		-1,   // BracedFieldValue
		-1,   // PositionalFieldName
		-1,   // PositionalFieldValue
		-1,   // FullSrec
		-1,   // OosvarValue
		-1,   // DirectOosvarValue
		-1,   // IndirectOosvarValue
		-1,   // BracedOosvarValue
		-1,   // FullOosvar
		-1,   // LocalVariable
		-1,   // Typedecl
		-1,   // Rvalue
		-1,   // PrecedenceChainStart
		-1,   // TernaryTerm
		-1,   // LogicalOrTerm
		-1,   // LogicalXORTerm
		-1,   // LogicalAndTerm
		-1,   // AbsentCoalesceTerm
		-1,   // EmptyCoalesceTerm
		-1,   // EqneTerm
		-1,   // CmpTerm
		-1,   // BitwiseORTerm
		-1,   // BitwiseXORTerm
		-1,   // BitwiseANDTerm
		-1,   // BitwiseShiftTerm
		-1,   // AddsubdotTerm
		-1,   // MuldivTerm
		-1,   // DotTerm
		-1,   // UnaryOpTerm
		-1,   // PowTerm
		-1,   // PrecedenceChainEnd
		-1,   // MlrvalOrFunction
		-1,   // ArrayLiteral
		-1,   // ArrayLiteralElements
		-1,   // MapLiteral
		-1,   // MapLiteralKeyValuePairs
		-1,   // MapLiteralKeyValuePair
		-1,   // ContextVariable
		-1,   // ENV
		-1,   // ArrayOrMapIndexAccess
		-1,   // ArrayOrMapPositionalNameAccess
		-1,   // ArrayOrMapPositionalValueAccess
		-1,   // ArraySliceAccess
		-1,   // FunctionCallsite
		-1,   // FunctionName
		-1,   // FcnArgs
		-1,   // SubroutineCallsite
		-1,   // SubroutineName
		-1,   // BracefulStatement
		-1,   // BeginBlock
		-1,   // EndBlock
		-1,   // CondBlock
		-1,   // IfChain
		-1,   // IfElifStar
		-1,   // IfBlock
		-1,   // ElifBlock
		-1,   // ElseBlock
		-1,   // WhileLoop
		-1,   // DoWhileLoop
		-1,   // ForLoop
		-1,   // ForLoopOneVariable
		-1,   // ForLoopTwoVariable
		-1,   // ForLoopMultivariable
		-1,   // MultiIndex
		-1,   // TripleForLoop
		-1,   // TripleForStart
		-1,   // TripleForContinuation
		-1,   // TripleForContinuationItem
		-1,   // TripleForUpdate
		-1,   // BreakStatement
		-1,   // ContinueStatement
		-1,   // NamedFunctionDefinition
		-1,   // UnnamedFunctionDefinition
		-1,   // SubroutineDefinition
		-1,   // FuncOrSubrParameterList
		-1,   // FuncOrSubrNonEmptyParameterList
		-1,   // FuncOrSubrParameter
		-1,   // UntypedFuncOrSubrParameterName
		-1,   // TypedFuncOrSubrParameterName
		-1,   // ReturnStatement
	},
	gotoRow{ // S4674
		-1, // S'
		-1, // Root
		-1, // StatementBlock
		-1, // NonEmptyStatementBlock
		-1, // StatementBlockInBraces
		-1, // BracelessStatement
		-1, // Assignment
		-1, // Unset
		-1, // Lvalue
		-1, // BareBoolean
		-1, // FilterStatement
		-1, // Redirector
		-1, // RedirectTarget
		-1, // PrintStatement
		-1, // PrintnStatement
		-1, // EprintStatement
		-1, // EprintnStatement
		-1, // DumpStatement
		-1, // EdumpStatement
		-1, // TeeStatement
		-1, // EmitFStatement
		-1, // Emit1Statement
		-1, // EmitStatement
		-1, // EmitPStatement
		-1, // EmittableList
		-1, // EmittableAsList
		-1, // Emittable
		-1, // EmitKeys
		-1, // FieldValue
		-1, // DirectFieldValue
		-1, // IndirectFieldValue
		-1, // BracedFieldValue
		-1, // PositionalFieldName
		-1, // PositionalFieldValue
		-1, // FullSrec
		-1, // OosvarValue
		-1, // DirectOosvarValue
		-1, // IndirectOosvarValue
		-1, // BracedOosvarValue
		-1, // FullOosvar
		-1, // LocalVariable
		-1, // Typedecl
		-1, // Rvalue
		-1, // PrecedenceChainStart
		-1, // TernaryTerm
		-1, // LogicalOrTerm
		-1, // LogicalXORTerm
		-1, // LogicalAndTerm
		-1, // AbsentCoalesceTerm
		-1, // EmptyCoalesceTerm
		-1, // EqneTerm
		-1, // CmpTerm
		-1, // BitwiseORTerm
		-1, // BitwiseXORTerm
		-1, // BitwiseANDTerm
		-1, // BitwiseShiftTerm
		-1, // AddsubdotTerm
		-1, // MuldivTerm
		-1, // DotTerm
		-1, // UnaryOpTerm
		-1, // PowTerm
		-1, // PrecedenceChainEnd
		-1, // MlrvalOrFunction
		-1, // ArrayLiteral
		-1, // ArrayLiteralElements
		-1, // MapLiteral
		-1, // MapLiteralKeyValuePairs
		-1, // MapLiteralKeyValuePair
		-1, // ContextVariable
		-1, // ENV
		-1, // ArrayOrMapIndexAccess
		-1, // ArrayOrMapPositionalNameAccess
		-1, // ArrayOrMapPositionalValueAccess
		-1, // ArraySliceAccess
		-1, // FunctionCallsite
		-1, // FunctionName
		-1, // FcnArgs
		-1, // SubroutineCallsite
		-1, // SubroutineName
		-1, // BracefulStatement
		-1, // BeginBlock
		-1, // EndBlock
		-1, // CondBlock
		-1, // IfChain
		-1, // IfElifStar
		-1, // IfBlock
		-1, // ElifBlock
		-1, // ElseBlock
		-1, // WhileLoop
		-1, // DoWhileLoop
		-1, // ForLoop
		-1, // ForLoopOneVariable
		-1, // ForLoopTwoVariable
		-1, // ForLoopMultivariable
		-1, // MultiIndex
		-1, // TripleForLoop
		-1, // TripleForStart
		-1, // TripleForContinuation
		-1, // TripleForContinuationItem
		-1, // TripleForUpdate
		-1, // BreakStatement
		-1, // ContinueStatement
		-1, // NamedFunctionDefinition
		-1, // UnnamedFunctionDefinition
		-1, // SubroutineDefinition
		-1, // FuncOrSubrParameterList
		-1, // FuncOrSubrNonEmptyParameterList
		-1, // FuncOrSubrParameter
		-1, // UntypedFuncOrSubrParameterName
		-1, // TypedFuncOrSubrParameterName
		-1, // ReturnStatement
	},
}
